import logging
import re
from string import Template
from typing import List

from kag.common.base.prompt_op import PromptOp

logger = logging.getLogger(__name__)

from kag.common.base.prompt_op import PromptOp


class LogicFormPlanPrompt(PromptOp):
    instruct_zh = """"instruction": "",
    "function_description": "functionName为算子名;基本格式为 functionName(arg_name1=arg_value1,[args_name2=arg_value2, args_name3=arg_value3]),括号中为参数，被[]包含的参数为可选参数，未被[]包含的为必选参数",
    "function": [
      {
          "functionName": "get_spo",
          "function_declaration": "get_spo(s=s_alias:entity_type[entity_name], p=p_alias:edge_type, o=o_alias:entity_type[entity_name], p.edge_type=value)",
          "description": "查找spo信息，s代表主体，o代表客体，表示为变量名:实体类型[实体名称]，实体名称作为可选参数，当有明确的查询实体时需要给出；p代表谓词，即关系或属性，表示为变量名:边类型或属性类型；这里为每个变量都分配一个变量名，作为后续提及时的指代；注意，s、p、o不能在同一表达式中反复多次出现；当变量为前文指代的变量名是，变量名必须和指代的变量名一致，且只需给出变量名，实体类型仅在首次引入时给定"
      },
      {
          "functionName": "count",
          "function_declaration": "count_alias=count(alias)",
          "description": "统计节点个数，参数为指定待统计的节点集合，只能是get_spo中出现的变量名；count_alias作为变量名表示计算结果，只能是int类型，变量名可作为下文的指代"
      },
      {
          "functionName": "sum",
          "function_declaration": "sum(alias, num1, num2, ...)->sum_alias",
          "description": "数据求和，参数为指定待求和的集合，可以是数字也可以是前文中出现的变量名，其内容只能是数值类型；sum_alias作为变量名表示计算结果，只能是数值类型，变量名可作为下文的指代"
      },
      {
          "functionName": "sort",
          "function_declaration": "sort(set=alias, orderby=o_alias or count_alias or sum_alias, direction=min or max, limit=N)",
          "description": "对节点集合排序，set指定待排序的节点集合，只能是get_spo中出现的变量名；orderby指定排序的依据，为节点的关系或属性名称，若是前文提及过的，则用别名指代；direction指定排序的方向，只能是min(正序)或max(倒序)排列；limit为输出个数限制，为int类型；可作为最后的输出结果"
      },
      {
          "functionName": "get",
          "function_decl:aration": "get(alias)",
          "description": "返回指定的别名代表的信息，可以是实体、关系路径或get_spo中获取到的属性值；可作为最后的输出结果"
      }
    ],
    """
    default_case_zh = """"cases": [
        {
            "Action": "吴京是谁",
            "answer": "Step1:查询吴京\nAction1:get_spo(s=s1:公众人物[吴京], p=p1, o=o1)\nOutput:输出s1\nAction2:get(s1)"
        },
        {
            "query": "30+6加上华为创始人在2024年的年龄是多少",
            "answer": "Step1:30+6 等于多少？\nAction1:sum(30,6)->sum1\nStep2:华为创始人是谁？\nAction2:get_spo(s=s2:企业[华为],p=p2:创始人,o=o2)\nStep3:华为创始人出生在什么年份？\nAction3:get_spo(s=o2,p=p3:出生年份,o=o3)\nStep4:华为创始人在2024年的年龄是多少？\nAction4:sum(2024,-o3)->sum4\nStep5:30+6的结果与华为创始人在2024年的年龄相加是多少？\nAction5:sum(sum1,sum4)->sum5\nStep6:输出sum5\nAction6:get(sum5)"
        }
    ],"""

    template_zh = f"""
{{
    {instruct_zh}
    {default_case_zh}
    "output_format": "only output `Step`, `Action` and `Output` content",
    "query": "$question"
}}   
    """

    instruct_en = """    "instruction": "",
    "function_description": "functionName is operator name;the function format is functionName(arg_name1=arg_value1,[args_name2=arg_value2, args_name3=arg_value3]),括号中为参数，被[]包含的参数为可选参数，未被[]包含的为必选参数",
    "function": [
      {
          "functionName": "get_spo",
          "function_declaration": "get_spo(s=s_alias:entity_type[entity_name], p=p_alias:edge_type, o=o_alias:entity_type[entity_name])",
          "description": "Find SPO information. 's' represents the subject, 'o' represents the object, and they are denoted as variable_name:entity_type[entity_name]. The entity name is an optional parameter and should be provided when there is a specific entity to query. 'p' represents the predicate, which can be a relationship or attribute, denoted as variable_name:edge_type_or_attribute_type. Each variable is assigned a unique variable name, which is used for reference in subsequent mentions. Note that 's', 'p', and 'o' should not appear repeatedly within the same expression; only one set of SPO should be queried at a time. When a variable is a reference to a previously mentioned variable name, the variable name must match the previously mentioned variable name, and only the variable name needs to be provided; the entity type is only given when it is first introduced."
      },
      {
          "functionName": "count",
          "function_declaration": "count(alias)->count_alias",
          "description": "Count the number of nodes. The parameter should be a specified set of nodes to count, and it can only be variable names that appear in the get_spo query. The variable name 'count_alias' represents the counting result, which must be of int type, and this variable name can be used for reference in subsequent mentions."
      },
      {
          "functionName": "sum",
          "function_declaration": "sum(alias, num1, num2, ...)->sum_alias",
          "description": "Calculate the sum of data. The parameter should be a specified set to sum, which can be either numbers or variable names mentioned earlier, and its content must be of numeric type. The variable name 'sum_alias' represents the result of the calculation, which must be of numeric type, and this variable name can be used for reference in subsequent mentions."      },
      {
          "functionName": "sort",
          "function_declaration": "sort(set=alias, orderby=o_alias or count_alias or sum_alias, direction=min or max, limit=N)",
          "description": "Sort a set of nodes. The 'set' parameter specifies the set of nodes to be sorted and can only be variable names that appear in the get_spo query. The 'orderby' parameter specifies the basis for sorting, which can be the relationship or attribute name of the nodes. If it has been mentioned earlier, an alias should be used. The 'direction' parameter specifies the sorting order, which can only be 'min' (ascending) or 'max' (descending). The 'limit' parameter specifies the limit on the number of output results and must be of int type. The sorted result can be used as the final output."      },
      {
          "functionName": "compare",
          "function_declaration": "compare(set=[alias1, alias2, ...], op=min|max)",
          "description": "Compare nodes or numeric values. The 'set' parameter specifies the set of nodes or values to be compared, which can be variable names that appear in the get_spo query or constants. The 'op' parameter specifies the comparison operation: 'min' to find the smallest and 'max' to find the largest."
      },
      {
          "functionName": "get",
          "function_decl:aration": "get(alias)",
          "description": "Return the information represented by a specified alias. This can be an entity, a relationship path, or an attribute value obtained in the get_spo query. It can be used as the final output result."
      }
    ],"""
    default_case_en = """"cases": [
        {
            "query": "Which sports team for which Cristiano Ronaldo played in 2011 was founded last ?",
            "answer": "Step1:Which Sports Teams Cristiano Ronaldo Played for in 2011 ?\nAction1:get_spo(s=s1:Player[Cristiano Ronaldo],p=p1:PlayedForIn2011Year,o=o1:SportsTeam)\nStep2:In which year were these teams established ?\nAction2:get_spo(s=o1,p=p2:FoundationYear,o=o2:Year)\nStep3:Which team was founded last ?\nAction3:sort(set=o1, orderby=o2, direction=max, limit=1)"
        },
        {
            "query": "Who was the first president of the association which published Journal of Psychotherapy Integration?",
            "answer": "Step1:Which association that publishes the Journal of Psychotherapy Integration ?\nAction1:Journal(s=s1:Player[Psychotherapy Integration],p=p1:Publish,o=o1:Association)\nStep2:Who was the first president of that specific association?\nAction2:get_spo(s=o1,p=p2:FirstPresident,o=o2:Person)"
        },
        {
            "query": "When did the state where Pocahontas Mounds is located become part of the United States?",
            "answer": "Step1:Which State Where Pocahontas Mounds is Located ?\nAction1:get_spo(s=s1:HistoricalSite[Pocahontas Mounds], p=p1:LocatedIn, o=o1:State)\nStep2:When did this state become a part of the United States ？\nAction2:get_spo(s=o1, p=p2:YearOfBecamingPartofTheUnitedStates, o=o2:Date)"
        },
        {
            "query": "Which of the two tornado outbreaks killed the most people?",
            "answer": "Step1:Which is the first tornado outbreaks ?\nAction1:get_spo(s=s1:Event[Tornado Outbreak], p=p1:TheFirst, o=o1:Event)\nStep2:Which is the second tornado outbreaks ?\nAction2:get_spo(s=s2:Event[Tornado Outbreak], p=p2:TheSecond, o=o2:Event)\nStep3:How many people died in the first tornado outbreak ?\nAction3:get_spo(s=s1, p=p3:KilledPeopleNumber, o=o3:Number)\nStep4:How many people died in the second tornado outbreak ?\nAction4:get_spo(s=s2, p=p4:KilledPeopleNumber, o=o4:Number)\nStep5:To compare the death toll between two tornado outbreaks to determine which one had more fatalities.\nAction5:compare(set=[o3,o4], op=max)"
        }
    ],"""
    template_en = f"""
{{
    {instruct_en}
    {default_case_en}
    "output_format": "Only output words in answer, for examples: `Step`, `Action` content",
    "query": "$question"
}}   
    """

    def __init__(self, language: str):
        super().__init__(language)

    @property
    def template_variables(self) -> List[str]:
        return ["question"]


    def parse_response(self, response: str, **kwargs):
        try:
            logger.debug(f"logic form:{response}")
            _output_string = response.replace("：", ":")
            _output_string = response.strip()
            sub_querys = []
            logic_forms = []
            current_sub_query = ''
            for line in _output_string.split('\n'):
                if line.startswith('Step'):
                    sub_querys_regex = re.search('Step\d+:(.*)', line)
                    if sub_querys_regex is not None:
                        sub_querys.append(sub_querys_regex.group(1))
                        current_sub_query = sub_querys_regex.group(1)
                elif line.startswith('Output'):
                    sub_querys.append("output")
                elif line.startswith('Action'):
                    logic_forms_regex = re.search('Action\d+:(.*)', line)
                    if logic_forms_regex:
                        logic_forms.append(logic_forms_regex.group(1))
                        if len(logic_forms) - len(sub_querys) == 1:
                            sub_querys.append(current_sub_query)
            return sub_querys, logic_forms
        except Exception as e:
            logger.warning(f"{response} parse logic form faied {e}", exc_info=True)
            return [], []



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

from typing import List, Type, Dict

from kag.interface.builder import AlignerABC
from knext.schema.client import BASIC_TYPES
from kag.builder.model.spg_record import SPGRecord
from kag.builder.model.sub_graph import SubGraph
from knext.common.base.runnable import Input, Output
from knext.schema.client import SchemaClient
from knext.schema.model.base import ConstraintTypeEnum, BaseSpgType


class SPGPostProcessorAligner(AlignerABC):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.spg_types = SchemaClient(project_id=self.project_id).load()

    @property
    def input_types(self) -> Type[Input]:
        return SPGRecord

    @property
    def output_types(self) -> Type[Output]:
        return SubGraph

    def merge(self, spg_records: List[SPGRecord]):
        merged_spg_records = {}
        for record in spg_records:
            key = f"{record.spg_type_name}#{record.get_property('name', '')}"
            if key not in merged_spg_records:
                merged_spg_records[key] = record
            else:
                old_record = merged_spg_records[key]
                for prop_name, prop_value in record.properties.items():
                    if prop_name not in old_record.properties:
                        old_record.properties[prop_name] = prop_value
                    else:
                        prop = self.spg_types.get(record.spg_type_name).properties.get(
                            prop_name
                        )
                        if not prop:
                            continue
                        if (
                            prop.object_type_name not in BASIC_TYPES
                            or prop.constraint.get(ConstraintTypeEnum.MultiValue)
                        ):
                            old_value = old_record.properties.get(prop_name)
                            if not prop_value:
                                prop_value = ""
                            prop_value_list = (
                                prop_value + "," + old_value
                                if old_value
                                else prop_value
                            ).split(",")
                            old_record.properties[prop_name] = ",".join(
                                list(set(prop_value_list))
                            )
                        else:
                            old_record.properties[prop_name] = prop_value

        return list(merged_spg_records.values())

    @staticmethod
    def from_spg_record(
        spg_types: Dict[str, BaseSpgType], spg_records: List[SPGRecord]
    ):
        sub_graph = SubGraph([], [])
        for record in spg_records:
            s_id = record.id
            s_name = record.name
            s_label = record.spg_type_name
            properties = record.properties

            spg_type = spg_types.get(record.spg_type_name)
            for prop_name, prop_value in record.properties.items():
                if prop_name in spg_type.properties:
                    from knext.schema.model.property import Property

                    prop: Property = spg_type.properties.get(prop_name)
                    o_label = prop.object_type_name
                    if o_label not in BASIC_TYPES:
                        prop_value_list = prop_value.split(",")
                        for o_id in prop_value_list:
                            sub_graph.add_edge(
                                s_id=s_id,
                                s_label=s_label,
                                p=prop_name,
                                o_id=o_id,
                                o_label=o_label,
                            )
                        properties.pop(prop_name)
            sub_graph.add_node(
                id=s_id, name=s_name, label=s_label, properties=properties
            )

        return sub_graph

    def invoke(self, input: Input, **kwargs) -> List[Output]:
        subgraph = SubGraph.from_spg_record(self.spg_types, [input])
        return [subgraph]

    def batch(self, inputs: List[Input], **kwargs) -> List[Output]:
        merged_records = self.merge(inputs)
        subgraph = SubGraph.from_spg_record(self.spg_types, merged_records)
        return [subgraph]



# coding=utf8
import os
import time
from typing import List

from knext.project.client import ProjectClient

from kag.common.vectorizer import Vectorizer
from kag.interface.retriever.kg_retriever_abc import KGRetrieverABC
from knext.search.client import SearchClient
from kag.solver.logic.core_modules.common.base_model import SPOEntity
from kag.solver.logic.core_modules.common.one_hop_graph import KgGraph, OneHopGraphData, EntityData
from kag.solver.logic.core_modules.common.schema_utils import SchemaUtils
from kag.solver.logic.core_modules.common.text_sim_by_vector import TextSimilarity
from kag.solver.logic.core_modules.common.utils import get_recall_node_label, generate_biz_id_with_type
from kag.solver.logic.core_modules.config import LogicFormConfiguration
from kag.solver.logic.core_modules.parser.logic_node_parser import GetSPONode, ParseLogicForm
from kag.solver.logic.core_modules.retriver.graph_retriver.dsl_executor import DslRunner, DslRunnerOnGraphStore
from kag.solver.logic.core_modules.retriver.retrieval_spo import FuzzyMatchRetrievalSpo, ExactMatchRetrievalSpo

current_dir = os.path.dirname(os.path.abspath(__file__))
import logging

logger = logging.getLogger()


class KGRetrieverByLlm(KGRetrieverABC):
    """
    A subclass of KGRetrieval that implements relation and entity retrieval using large language models.

    This class provides the default implementation for retrieving relations and entities within the system,
    leveraging large language models for its operations.
    """

    def __init__(self, disable_exact_match=False, **kwargs):
        super().__init__(**kwargs)

        vectorizer_config = eval(os.getenv("KAG_VECTORIZER", "{}"))
        if self.host_addr and self.project_id:
            config = ProjectClient(host_addr=self.host_addr, project_id=self.project_id).get_config(self.project_id)
            vectorizer_config.update(config.get("vectorizer", {}))
        self.vectorizer: Vectorizer = Vectorizer.from_config(vectorizer_config)
        self.text_similarity = TextSimilarity(vec_config=vectorizer_config)
        self.schema = SchemaUtils(LogicFormConfiguration(kwargs))
        self.schema.get_schema()

        self.disable_exact_match = disable_exact_match

        self.sc: SearchClient = SearchClient(self.host_addr, self.project_id)
        self.dsl_runner: DslRunner = DslRunnerOnGraphStore(self.project_id, self.schema, LogicFormConfiguration(kwargs))

        self.fuzzy_match = FuzzyMatchRetrievalSpo(text_similarity=self.text_similarity, llm=self.llm_module)
        self.exact_match = ExactMatchRetrievalSpo(self.schema)
        self.parser = ParseLogicForm(self.schema, None)

    def retrieval_relation(self, n: GetSPONode, one_hop_graph_list: List[OneHopGraphData], **kwargs) -> KgGraph:
        req_id = kwargs.get('req_id', '')
        debug_info = kwargs.get('debug_info', {})
        if not self.disable_exact_match:
            process_kg, is_matched = self._exact_match_spo(n, one_hop_graph_list, req_id)
            if is_matched:
                debug_info["exact_match_spo"] = True and debug_info.get('exact_match_spo', True)
                return process_kg
            else:
                debug_info["exact_match_spo"] = False
        return self._fuzzy_match_spo(n, one_hop_graph_list, req_id)

    def retrieval_entity(self, mention_entity: SPOEntity, topk=5, **kwargs) -> List[EntityData]:
        recalled_el_set = self._search_retrieval_entity(mention_entity, topk=topk, kwargs=kwargs)
        if len(mention_entity.value_list) == 0:
            return recalled_el_set
        # 存在参数进行过滤，先过去一跳子图
        one_hop_graph_map = self.dsl_runner.query_vertex_one_graph_by_s_o_ids(recalled_el_set, [], {})
        matched_entity_list = recalled_el_set
        # 将待匹配的作为spg进行匹配
        for k, v in mention_entity.value_list:
            choosed_one_hop_graph_list = self._get_matched_one_hop(one_hop_graph_map, matched_entity_list)
            param_spo = f"get_spo(s=s1:{mention_entity.get_entity_first_type_or_zh()}[{mention_entity.entity_name}],p=p1:{k},o=o1:Entity[{v}])"
            tmp_spo = self.parser.parse_logic_form(param_spo, parsed_entity_set={}, sub_query=f"{mention_entity.entity_name} {k} {v}")
            debug_info = {}
            kg_graph = self.retrieval_relation(tmp_spo, choosed_one_hop_graph_list, debug_info=debug_info)
            kg_graph.nodes_alias.append(tmp_spo.s.alias_name)
            kg_graph.nodes_alias.append(tmp_spo.o.alias_name)
            kg_graph.edge_alias.append(tmp_spo.p.alias_name)
            matched_entity_list = kg_graph.get_entity_by_alias("s1")
            if matched_entity_list is None:
                return []

        return matched_entity_list

    def _get_matched_one_hop(self, one_hop_graph_map: dict, matched_entity_list: list):
        ret_one_hop_list = []
        for matched_entity in matched_entity_list:
            cached_id = generate_biz_id_with_type(matched_entity.biz_id,
                                                  matched_entity.type_zh if matched_entity.type_zh else matched_entity.type)
            if cached_id in one_hop_graph_map:
                ret_one_hop_list.append(one_hop_graph_map[cached_id])
        return ret_one_hop_list

    def _search_retrieval_entity(self, mention_entity: SPOEntity, topk=5, **kwargs) -> List[EntityData]:
        retdata = []
        if mention_entity is None:
            return retdata
        content = kwargs.get('content', mention_entity.entity_name)
        query_type = mention_entity.get_entity_first_type_or_zh()
        recognition_threshold = kwargs.get('recognition_threshold', 0.8)
        recall_topk = topk
        if "entity" not in query_type.lower():
            recall_topk = 10
        query_vector = self.vectorizer.vectorize(mention_entity.entity_name)
        typed_nodes = self.sc.search_vector(
            label="Entity", property_key="name", query_vector=query_vector, topk=recall_topk
        )
        # 根据query召回
        if query_type not in ["Others", "Entity"]:
            content_vector = self.vectorizer.vectorize(content)
            content_recall_nodes = self.sc.search_vector(
                label="Entity", property_key="desc", query_vector=content_vector, topk=recall_topk
            )
        else:
            content_recall_nodes = []
        sorted_nodes = typed_nodes + content_recall_nodes
        if len(sorted_nodes) == 0:
            sorted_nodes = self.sc.search_text(query_string=mention_entity.entity_name)

        # rerank
        def rerank_sematic_type(cands_nodes: list, sematic_type: str):
            sematic_type_list = []
            for cands in cands_nodes:
                node = cands['node']
                if "semanticType" not in node.keys() or node['semanticType'] == '':
                    continue
                sematic_type_list.append(node['semanticType'])
            sematic_type_list = list(set(sematic_type_list))
            sematic_match_score_list = self.text_similarity.text_sim_result(sematic_type, sematic_type_list,
                                                                            len(sematic_type_list), low_score=-1)
            sematic_match_score_map = {}
            for i in sematic_match_score_list:
                sematic_match_score_map[i[0]] = i[1]
            for node in cands_nodes:
                recall_node_label = get_recall_node_label(node['node']['__labels__'])
                if recall_node_label == sematic_type:
                    node['type_match_score'] = node['score']
                elif "semanticType" not in node['node'].keys() or node['node']['semanticType'] == '':
                    node['type_match_score'] = 0.3
                else:
                    node['type_match_score'] = node['score'] * sematic_match_score_map[node['node']['semanticType']]
            sorted_people_dicts = sorted(cands_nodes, key=lambda node: node['type_match_score'], reverse=True)
            # 取top5
            return sorted_people_dicts[:topk]

        if "entity" not in query_type.lower():
            sorted_nodes = rerank_sematic_type(sorted_nodes, query_type)
        sorted_people_dicts = sorted(sorted_nodes, key=lambda node: node['score'], reverse=True)
        for recall in sorted_people_dicts:
            if len(sorted_people_dicts) != 0 and recall["score"] >= recognition_threshold:
                recalled_entity = EntityData()
                recalled_entity.score = recall["score"]
                recalled_entity.biz_id = recall["node"]["id"]
                recalled_entity.name = recall["node"]["name"]
                recalled_entity.type = get_recall_node_label(recall["node"]["__labels__"])
                retdata.append(recalled_entity)
            else:
                break
        return retdata[:topk]

    def _exact_match_spo(self, n: GetSPONode, one_hop_graph_list: List[OneHopGraphData], req_id: str):
        start_time = time.time()
        total_one_kg_graph, matched_flag = self.exact_match.match_spo(n, one_hop_graph_list)
        logger.debug(
            f"{req_id} _exact_match_spo cost={time.time() - start_time} matched_flag={matched_flag}")
        if not matched_flag:
            return total_one_kg_graph, matched_flag
        for alias_name in total_one_kg_graph.entity_map.keys():
            for e in total_one_kg_graph.entity_map[alias_name]:
                score = e.score
                if score < 0.9:
                    total_one_kg_graph.rmv_node_ins(alias_name, [e.biz_id])
                    return total_one_kg_graph, False
        return total_one_kg_graph, matched_flag

    def _fuzzy_match_spo(self, n: GetSPONode, one_hop_graph_list: List[OneHopGraphData], req_id: str):
        start_time = time.time()

        total_one_kg_graph, matched_flag = self.fuzzy_match.match_spo(n, one_hop_graph_list)
        logger.debug(
            f"{req_id} _fuzzy_match_spo cost={time.time() - start_time} matched_flag={matched_flag}")
        return total_one_kg_graph



import logging
from typing import List

from kag.common.base.prompt_op import PromptOp

logger = logging.getLogger(__name__)


class DeduceEntail(PromptOp):
    template_zh = "根据提供的选项及相关答案，请选择其中一个选项回答问题“$instruction”。" \
                  "无需解释；" \
                  "如果没有可选择的选项，直接回复“无相关信息”无需解释" \
                  "\n【信息】：“$memory”\n请确保所提供的信息直接准确地来自检索文档，不允许任何自身推测。"
    template_en = "Based on the provided options and related answers, choose one option to respond to the question '$instruction'." \
                  "No explanation is needed;" \
                  "If there are no available options, simply reply 'No relevant information' without explanation." \
                  "\n[Information]: '$memory'" \
                  "\nEnsure that the information provided comes directly and accurately from the retrieved document, " \
                  "without any speculation."


    def __init__(self, language: str):
        super().__init__(language)

    @property
    def template_variables(self) -> List[str]:
        return ["memory", "instruction"]

    def parse_response_en(self, satisfied_info: str):
        if satisfied_info.startswith('No relevant information'):
            if_answered = False
        else:
            if_answered = True
        return if_answered, satisfied_info

    def parse_response_zh(self, satisfied_info: str):
        if satisfied_info.startswith("无相关信息"):
            if_answered = False
        else:
            if_answered = True
        return if_answered, satisfied_info

    def parse_response(self, response: str, **kwargs):
        logger.debug('推理器判别:{}'.format(response))
        if self.language == "en":
            return self.parse_response_en(response)
        return self.parse_response_zh(response)



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import json
from string import Template
from typing import List, Optional

from kag.common.base.prompt_op import PromptOp
from knext.reasoner.client import ReasonerClient


class QuestionNER(PromptOp):

    template_en = """
    {
        "instruction": "You are an expert in named entity recognition. Please extract entities and that match the schema definition from the input. Return an empty list if the entity type does not exist. Please respond in the format of a JSON string.You can refer to the example for extraction.",
        "schema": $schema,
        "example": [
            {
                "input": "Which magazine was started first, Arthur's Magazine or First for Women?",
                "output": [
                        {
                            "entity": "First for Women",
                            "category": "Works"
                        },
                        {
                            "entity": "Arthur's Magazine",
                            "category": "Works"
                        }
                    ]
            }
        ],
        "input": "$input"
    }    
        """

    template_zh = template_en

    def __init__(
            self, language: Optional[str] = "en", **kwargs
    ):
        super().__init__(language, **kwargs)
        self.schema = ReasonerClient(project_id=self.project_id).get_reason_schema().keys()
        self.template = Template(self.template).safe_substitute(schema=self.schema)

    @property
    def template_variables(self) -> List[str]:
        return ["input"]

    def parse_response(self, response: str, **kwargs):
        rsp = response
        if isinstance(rsp, str):
            rsp = json.loads(rsp)
        if isinstance(rsp, dict) and "output" in rsp:
            rsp = rsp["output"]
        if isinstance(rsp, dict) and "named_entities" in rsp:
            entities = rsp["named_entities"]
        else:
            entities = rsp

        return entities



import logging
import os
import time
from typing import List

from kag.common.graphstore.graph_store import GraphStore
from kag.interface.retriever.chunk_retriever_abc import ChunkRetrieverABC
from kag.interface.retriever.kg_retriever_abc import KGRetrieverABC
from kag.solver.common.base import Question
from kag.solver.logic.core_modules.common.base_model import LFPlanResult
from kag.solver.logic.core_modules.common.one_hop_graph import KgGraph
from kag.solver.logic.core_modules.common.schema_utils import SchemaUtils
from kag.solver.logic.core_modules.common.text_sim_by_vector import TextSimilarity
from kag.solver.logic.core_modules.config import LogicFormConfiguration
from kag.solver.logic.core_modules.op_executor.op_deduce.deduce_executor import DeduceExecutor
from kag.solver.logic.core_modules.op_executor.op_math.math_executor import MathExecutor
from kag.solver.logic.core_modules.op_executor.op_output.output_executor import OutputExecutor
from kag.solver.logic.core_modules.op_executor.op_retrieval.retrieval_executor import RetrievalExecutor
from kag.solver.logic.core_modules.op_executor.op_sort.sort_executor import SortExecutor
from kag.solver.logic.core_modules.parser.logic_node_parser import ParseLogicForm
from kag.solver.logic.core_modules.retriver.entity_linker import EntityLinkerBase
from kag.solver.logic.core_modules.retriver.graph_retriver.dsl_executor import DslRunner, DslRunnerOnGraphStore
from kag.solver.logic.core_modules.retriver.schema_std import SchemaRetrieval
from kag.solver.logic.core_modules.rule_runner.rule_runner import OpRunner
from kag.solver.tools.info_processor import ReporterIntermediateProcessTool

logger = logging.getLogger()


class LogicExecutor:
    def __init__(self, query: str, project_id: str,
                 schema: SchemaUtils, kg_retriever: KGRetrieverABC,
                 chunk_retriever: ChunkRetrieverABC, std_schema: SchemaRetrieval, el: EntityLinkerBase, generator,
                 dsl_runner: DslRunner,
                 text_similarity: TextSimilarity=None,
                 req_id='',
                 need_detail=False, llm=None, report_tool=None, params=None):
        """
        Initializes the LogicEngine with necessary parameters and configurations.

        :param query: The main query to process.
        :param project_id: Identifier for the project.
        :param schema: The schema used for processing.
        :chunk_retriever (ChunkRetriever): An instance for chunk-level retrieval. If not provided, we will not execute chunk retrieval.
        :kg_retriever (KGRetrieval): An instance for graph-level retrieval. If not provided, we will not execute retrieval on graph.
        :param std_schema: Standard schema retrieval instance for retrieval schema label.
        :param el: Entity linker base instance.
        :param generator: Generator for generating answers.
        :param dsl_runner: DSL runner instance, can run cypher to query graph. Defaults to `None`.
        :param text_similarity: convert text to vector, and compute similarity score
        :param req_id: Request identifier. Defaults to an empty string.
        :param need_detail: Flag indicating whether detailed information is needed. Defaults to `False`.
        :param llm: Language model instance. Defaults to `None`.
        :param report_tool: Reporting tool instance. Defaults to `None`.
        :param params: Additional parameters. Defaults to `None`.
        """
        # pipeline record
        if params is None:
            params = {}
        self.report_tool = report_tool
        self.need_detail = need_detail
        self.req_id = req_id
        self.params = params
        self.dsl_runner = dsl_runner
        self.schema = schema
        self.project_id = project_id
        self.nl_query = query
        self.query_one_graph_cache = {}
        self.kg_graph = KgGraph()
        self.kg_retriever = kg_retriever
        self.chunk_retriever = chunk_retriever
        self.alias_entity = {}
        self.query_schema = {}
        self.need_query_one_hop_with_rel = []
        self.recall_data_time = 0
        self.debug_info = {
            "el": [],
            "el_detail": [],
            "std_out": [],
            "get_empty": [],
            "sub_qa_pair": []
        }
        self.op_runner = OpRunner(self.kg_graph, llm, query, self.req_id)
        self.parser = ParseLogicForm(self.schema, std_schema)
        self.text_similarity = text_similarity or TextSimilarity()
        self.llm = llm
        self.generator = generator
        self.el = el

        self.force_chunk_retriever = os.getenv("KAG_QA_FORCE_CHUNK_RETRIEVER", False)

        # Initialize executors for different operations.
        self.retrieval_executor = RetrievalExecutor(query, self.kg_graph, self.schema, self.kg_retriever,
                                                    self.el,
                                                    self.dsl_runner, self.debug_info, text_similarity,KAG_PROJECT_ID = self.project_id)
        self.deduce_executor = DeduceExecutor(query, self.kg_graph, self.schema, self.op_runner, self.debug_info, KAG_PROJECT_ID = self.project_id)
        self.sort_executor = SortExecutor(query, self.kg_graph, self.schema, self.debug_info, KAG_PROJECT_ID = self.project_id)
        self.math_executor = MathExecutor(query, self.kg_graph, self.schema, self.debug_info, KAG_PROJECT_ID = self.project_id)
        self.output_executor = OutputExecutor(query, self.kg_graph, self.schema, self.el,
                                              self.dsl_runner,
                                              self.retrieval_executor.query_one_graph_cache, self.debug_info, KAG_PROJECT_ID = self.project_id)

        self.with_sub_answer = os.getenv("KAG_QA_WITH_SUB_ANSWER", True)

    def _convert_logic_nodes_2_question(self, logic_nodes: List[LFPlanResult]) -> List[Question]:
        ret_question = []
        for i in range(0, len(logic_nodes)):
            if i == 0:
                question = Question(
                    question=logic_nodes[i].query,
                )
            else:
                question = Question(
                    question=logic_nodes[i].query,
                    dependencies=[ret_question[i - 1]]
                )
            question.id = i
            ret_question.append(question)
        return ret_question

    def _generate_sub_answer(self, history: list, spo_retrieved: list, docs: list, sub_query: str):
        if not self.with_sub_answer:
            return "I don't know"
        return self.generator.generate_sub_answer(sub_query, spo_retrieved, docs, history)

    def execute(self, lf_nodes: List[LFPlanResult], init_query):
        """
        Executes the logic nodes and processes the initial query to retrieve answers.

        :param lf_nodes: List of logic nodes to be executed.
        :param init_query: The initial query that triggered this execution.
        :return: A tuple containing the QA results, knowledge graph, and execution history.
        """
        self._create_report_pipeline(init_query, lf_nodes)

        kg_qa_result = []
        history = []
        query_ner_list = {}
        # get NER results for the initial query, for chunk retrieve
        if self.chunk_retriever and hasattr(self.chunk_retriever, 'get_std_ner_by_query'):
            query_ner_list = self.chunk_retriever.get_std_ner_by_query(init_query)

        query_num = 0

        # Process each sub-query.
        for lf in lf_nodes:
            sub_query, sub_logic_nodes = lf.query, lf.lf_nodes
            query_num += 1
            node_begin_time = time.time()
            sub_logic_nodes_str = "\n".join([str(ln) for ln in sub_logic_nodes])

            question = self._create_sub_question_report_node(query_num, sub_logic_nodes_str, sub_query)
            if self.kg_retriever:
                kg_qa_result, spo_retrieved = self._execute_lf(sub_logic_nodes)
            else:
                logger.info(f"lf executor disabled kg retriever {init_query}")
                kg_qa_result, spo_retrieved = [], []

            question.context.append(f"#### spo retrieved:")
            question.context.append(f"{spo_retrieved if len(spo_retrieved) > 0 else 'no spo tuple retrieved'}.")
            self._update_sub_question_status(question, None, ReporterIntermediateProcessTool.STATE.RUNNING)


            answer_source = "spo"
            docs_with_score = []
            all_related_entities, sub_answer = self._generate_sub_answer_by_graph(
                history, kg_qa_result, spo_retrieved, sub_query)

            # if sub answer is `I don't know`, we use chunk retriever
            if "i don't know" in sub_answer.lower() and self.chunk_retriever:
                answer_source = "chunk"
                question.context.append(f"## Chunk Retriever")
                self._update_sub_question_status(question, None, ReporterIntermediateProcessTool.STATE.RUNNING)

                start_time = time.time()
                # Update parameters to include retrieved SPO entities as starting points for chunk retrieval.
                params = {
                    'related_entities': all_related_entities,
                    'query_ner_dict': query_ner_list,
                    'req_id': self.req_id
                }
                # Retrieve chunks using the updated parameters.
                sub_query_with_history_qa = self._generate_sub_query_with_history_qa(history, sub_query)
                docs_with_score = self.chunk_retriever.recall_docs(sub_query_with_history_qa, top_k=10, **params)
                docs = ["#".join(item.split("#")[:-1]) for item in docs_with_score]

                self._update_sub_question_recall_docs(docs, question)
                self._update_sub_question_status(question, None, ReporterIntermediateProcessTool.STATE.RUNNING)

                retrival_time = time.time() - start_time
                sub_answer = self._generate_sub_answer(history, spo_retrieved, docs, sub_query)
                question.context.append("#### answer based by fuzzy retrieved:")
                question.context.append(f"{sub_answer}")
                logger.info(f"{self.req_id} call by docs cost: {retrival_time} docs num={len(docs)}")

            history.append(
                {"sub_query": sub_query, "sub_answer": sub_answer, 'docs': docs_with_score,
                 'spo_retrieved': spo_retrieved,
                 'exactly_match': self.debug_info.get('exact_match_spo', False),
                 'logic_expr': sub_logic_nodes_str, 'answer_source': answer_source,
                 'cost': time.time() - node_begin_time})
            self.debug_info['sub_qa_pair'].append([sub_query, sub_answer])
            self._update_sub_question_status(question, sub_answer, ReporterIntermediateProcessTool.STATE.FINISH)

        return kg_qa_result, self.kg_graph, history

    def _generate_sub_query_with_history_qa(self, history, sub_query):
        # Generate a sub-query with history qa pair
        if history:
            history_sub_answer = [h['sub_answer'] for h in history[:3] if
                                  "i don't know" not in h['sub_answer'].lower()]
            sub_query_with_history_qa = '\n'.join(history_sub_answer) + '\n' + sub_query
        else:
            sub_query_with_history_qa = sub_query
        return sub_query_with_history_qa

    def _update_sub_question_recall_docs(self, docs, question):
        question.context.extend(["|id|content|", "|-|-|"])
        for i, d in enumerate(docs, start=1):
            _d = d.replace('\n', '<br>')
            question.context.append(f"|{i}|{_d}|")

    def _generate_sub_answer_by_graph(self, history, kg_qa_result, spo_retrieved, sub_query):
        sub_answer = "I don't know"
        all_related_entities = self.kg_graph.get_all_entity()
        all_related_entities = list(set(all_related_entities))
        if self.force_chunk_retriever:
            # if this flag is true, we force use chunk retriever
            logger.info(f"lf executor disabled with_sub_answer {sub_query}")
            return all_related_entities, sub_answer
        # if spo retrieved empty, we use chunk retriever to retrieve answer
        if len(spo_retrieved) == 0 and len(kg_qa_result) == 0:
            all_related_entities = []
        # if there is answer in kg_qa_result, and the answer is exact match with spo, we generate answer with kg result
        elif self.debug_info.get('exact_match_spo', False):
            if len(kg_qa_result) > 0:
                sub_answer = str(kg_qa_result)
            else:
                sub_answer = str(spo_retrieved)
        # other condition, we generate answer by graph, if can not generate answer, we will return `I don't know`
        else:
            if len(spo_retrieved) == 0:
                spo_retrieved = kg_qa_result
            sub_answer = self._generate_sub_answer(history, spo_retrieved, [], sub_query)
        return all_related_entities, sub_answer

    def _update_sub_question_status(self, question, answer, status):
        if self.report_tool:
            self.report_tool.report_node(question, answer, status)

    def _create_sub_question_report_node(self, query_num, sub_logic_nodes_str, sub_query):
        question = Question(
            question=sub_query,
        )
        question.id = query_num
        question.context = ["## SPO Retriever", "#### logic_form expression: ",
                            f'```java\n{sub_logic_nodes_str}\n```']
        if self.report_tool:
            self.report_tool.report_node(question, None, ReporterIntermediateProcessTool.STATE.RUNNING)
        return question

    def _create_report_pipeline(self, init_query, lf_nodes):
        if self.report_tool:
            self.report_tool.report_pipeline(Question(question=init_query), self._convert_logic_nodes_2_question(lf_nodes))

    def _execute_lf(self, sub_logic_nodes):
        kg_qa_result = []
        spo_set = []
        # Execute graph retrieval operations.
        for n in sub_logic_nodes:
            logger.debug(f"{self.req_id} begin run logic node " + str(n))
            if self.retrieval_executor.is_this_op(n):
                self.retrieval_executor.executor(n, self.req_id, self.params)
                cur_spo_set = self.kg_graph.get_entity_by_alias(n.p.alias_name)
                if cur_spo_set is not None and len(cur_spo_set) > 0:
                    spo_set += [str(spo) for spo in cur_spo_set]
            elif self.deduce_executor.is_this_op(n):
                deduce_res = self.deduce_executor.executor(n, self.req_id, self.params)
                if isinstance(deduce_res, list):
                    kg_qa_result += deduce_res
            elif self.math_executor.is_this_op(n):
                self.math_executor.executor(n, self.req_id, self.params)
            elif self.sort_executor.is_this_op(n):
                self.sort_executor.executor(n, self.req_id, self.params)
            elif self.output_executor.is_this_op(n):
                kg_qa_result += self.output_executor.executor(n, self.req_id, self.params)
            else:
                logger.warning(f"unknown operator: {n.operator}")
        return kg_qa_result, spo_set



import re
from string import Template
from typing import List
import logging

from kag.common.base.prompt_op import PromptOp

logger = logging.getLogger(__name__)


class RespJudge(PromptOp):
    template_zh = "根据当前已知信息进行判断，不允许进行推理，" \
                  "你能否完全并准确地回答这个问题'$instruction'?\n已知信息：'$memory'。" \
                  "\n如果你能，请直接回复‘是’\n如果不能且需要更多信息，请直接回复‘否’。"
    template_en = "Judging based solely on the current known information and without allowing for inference, " \
                  "are you able to completely and accurately respond to the question '$instruction'? " \
                  "\nKnown information: '$memory'. " \
                  "\nIf you can, please reply with 'Yes' directly; " \
                  "if you cannot and need more information, please reply with 'No' directly."

    def __init__(self, language: str):
        super().__init__(language)

    @property
    def template_variables(self) -> List[str]:
        return ["memory", "instruction"]

    def parse_response_en(self, satisfied_info: str):
        if satisfied_info[:3] == 'Yes':
            if_finished = True
        else:
            if_finished = False
        return if_finished

    def parse_response_zh(self, satisfied_info: str):
        if satisfied_info.startswith("是"):
            if_finished = True
        else:
            if_finished = False
        return if_finished

    def parse_response(self, response: str, **kwargs):
        logger.debug('推理器判别:{}'.format(response))
        if self.language == "en":
            return self.parse_response_en(response)
        return self.parse_response_zh(response)



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.
import copy
import logging
from typing import List, Dict

from tenacity import retry, stop_after_attempt

from kag.builder.component.extractor import KAGExtractor
from kag.builder.model.sub_graph import SubGraph
from kag.builder.prompt.spg_prompt import SPG_KGPrompt
from kag.common.base.prompt_op import PromptOp
from knext.common.base.runnable import Input, Output

from knext.schema.client import BASIC_TYPES

logger = logging.getLogger(__name__)


class SPGExtractor(KAGExtractor):
    """
    A Builder Component that extracting structured data from long texts by invoking large language model.

    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.spg_ner_types, self.kag_ner_types = [], []
        for type_name, spg_type in self.schema.items():
            properties = list(spg_type.properties.keys())
            for p in properties:
                if p not in SPG_KGPrompt.ignored_properties:
                    self.spg_ner_types.append(type_name)
                    continue
            self.kag_ner_types.append(type_name)
        self.kag_ner_prompt = PromptOp.load(self.biz_scene, "ner")(language=self.language, project_id=self.project_id)
        self.spg_ner_prompt = SPG_KGPrompt(self.spg_ner_types, self.language)

    @retry(stop=stop_after_attempt(3))
    def named_entity_recognition(self, passage: str):
        """
        Performs named entity recognition on a given text passage.
        Args:
            passage (str): The text to perform named entity recognition on.
        Returns:
            The result of the named entity recognition operation.
        """
        spg_ner_result = self.llm.batch({"input": passage}, self.spg_ner_prompt)
        kag_ner_result = self.llm.invoke({"input": passage}, self.kag_ner_prompt)
        return spg_ner_result + kag_ner_result

    def assemble_sub_graph_with_spg_records(self, entities: List[Dict]):
        sub_graph = SubGraph([], [])
        for record in entities:
            s_name = record.get("entity", "")
            s_label = record.get("category", "")
            properties = record.get("properties", {})
            tmp_properties = copy.deepcopy(properties)
            spg_type = self.schema.get(s_label)
            for prop_name, prop_value in properties.items():
                if prop_value == "NAN":
                    tmp_properties.pop(prop_name)
                    continue
                if prop_name in spg_type.properties:
                    from knext.schema.model.property import Property
                    prop: Property = spg_type.properties.get(prop_name)
                    o_label = prop.object_type_name_en
                    if o_label not in BASIC_TYPES:
                        if isinstance(prop_value, str):
                            prop_value = [prop_value]
                        for o_name in prop_value:
                            sub_graph.add_node(id=o_name, name=o_name, label=o_label)
                            sub_graph.add_edge(s_id=s_name, s_label=s_label, p=prop_name, o_id=o_name, o_label=o_label)
                        tmp_properties.pop(prop_name)
            record["properties"] = tmp_properties
            sub_graph.add_node(id=s_name, name=s_name, label=s_label, properties=properties)
        return sub_graph, entities

    def invoke(self, input: Input, **kwargs) -> List[Output]:
        """
        Invokes the semantic extractor to process input data.

        Args:
            input (Input): Input data containing name and content.
            **kwargs: Additional keyword arguments.

        Returns:
            List[Output]: A list of processed results, containing subgraph information.
        """
        title = input.name
        passage = title + "\n" + input.content

        try:
            entities = self.named_entity_recognition(passage)
            sub_graph, entities = self.assemble_sub_graph_with_spg_records(entities)
            filtered_entities = [{k: v for k, v in ent.items() if k in ["entity", "category"]} for ent in entities]
            triples = self.triples_extraction(passage, filtered_entities)
            std_entities = self.named_entity_standardization(passage, filtered_entities)
            self.append_official_name(entities, std_entities)
            self.assemble_sub_graph(sub_graph, input, entities, triples)
            return [sub_graph]
        except Exception as e:
            import traceback

            traceback.print_exc()
            logger.info(e)
        return []



import logging

from kag.interface.solver.kag_generator_abc import KAGGeneratorABC
from kag.interface.solver.kag_reasoner_abc import KagReasonerABC
from kag.interface.solver.kag_reflector_abc import KagReflectorABC
from kag.solver.implementation.default_generator import DefaultGenerator
from kag.solver.implementation.default_memory import DefaultMemory
from kag.solver.implementation.default_reasoner import DefaultReasoner
from kag.solver.implementation.default_reflector import DefaultReflector

logger = logging.getLogger(__name__)


class SolverPipeline:
    def __init__(self, max_run=3, reflector: KagReflectorABC = None, reasoner: KagReasonerABC = None,
                 generator: KAGGeneratorABC = None, **kwargs):
        """
        Initializes the think-and-act loop class.

        :param max_run: Maximum number of runs to limit the thinking and acting loop, defaults to 3.
        :param reflector: Reflector instance for reflect tasks.
        :param reasoner: Reasoner instance for reasoning about tasks.
        :param generator: Generator instance for generating actions.
        """
        self.max_run = max_run
        self.memory = DefaultMemory(**kwargs)

        self.reflector = reflector or DefaultReflector(**kwargs)
        self.reasoner = reasoner or DefaultReasoner(**kwargs)
        self.generator = generator or DefaultGenerator(**kwargs)

        self.trace_log = []

    def run(self, question):
        """
       Executes the core logic of the problem-solving system.

       Parameters:
       - question (str): The question to be answered.

       Returns:
       - tuple: answer, trace log
       """
        instruction = question
        if_finished = False
        logger.debug('input instruction:{}'.format(instruction))
        present_instruction = instruction
        run_cnt = 0

        while not if_finished and run_cnt < self.max_run:
            run_cnt += 1
            logger.debug('present_instruction is:{}'.format(present_instruction))
            # Attempt to solve the current instruction and get the answer, supporting facts, and history log
            solved_answer, supporting_fact, history_log = self.reasoner.reason(present_instruction)

            # Extract evidence from supporting facts
            self.memory.save_memory(solved_answer, supporting_fact, instruction)

            history_log['present_instruction'] = present_instruction
            history_log['present_memory'] = self.memory.serialize_memory()
            self.trace_log.append(history_log)

            # Reflect the current instruction based on the current memory and instruction
            if_finished, present_instruction = self.reflector.reflect_query(self.memory, present_instruction)

        response = self.generator.generate(instruction, self.memory)
        return response, self.trace_log

    def get_kg_answer_num(self):
        """
        Get the number of direct answers from the knowledge graph.
        Debug Info

        Returns:
            int: The number of direct answers from the knowledge graph.
        """
        return self.reasoner.kg_direct

    def get_total_sub_question_num(self):
        """
        Get the total number of sub-questions for solve question.
        Debug Info

        Returns:
            int: The total number of sub-questions.
        """
        return self.reasoner.sub_query_total



# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

from abc import ABC, abstractmethod


class GraphStore(ABC):
    """
    Abstract base class for a graph store that defines standard interfaces for graph data operations.
    This class specifies abstract methods to ensure subclasses implement specific graph operations such as node CRUD, relationship handling, and index management.
    """

    @abstractmethod
    def close(self):
        """
        Close the graph store resources.
        """
        pass

    @abstractmethod
    def initialize_schema(self, schema):
        """
        Initialize the graph schema.

        Parameters:
        - schema: Definition of the graph schema.
        """
        pass

    @abstractmethod
    def upsert_node(self, label, properties, id_key="id", extra_labels=("Entity",)):
        """
        Insert or update a single node.

        Parameters:
        - label: Label of the node.
        - properties: Properties of the node.
        - id_key: Property key used as the unique identifier.
        - extra_labels: Additional labels for the node.
        """
        pass

    @abstractmethod
    def upsert_nodes(self, label, properties_list, id_key="id", extra_labels=("Entity",)):
        """
        Insert or update multiple nodes.

        Parameters:
        - label: Label of the nodes.
        - properties_list: List of properties for the nodes.
        - id_key: Property key used as the unique identifier.
        - extra_labels: Additional labels for the nodes.
        """
        pass

    @abstractmethod
    def batch_preprocess_node_properties(self, node_batch, extra_labels=("Entity",)):
        """
        Batch preprocess node properties.

        Parameters:
        - node_batch: A batch of nodes.
        - extra_labels: Additional labels for the nodes.
        """
        pass

    @abstractmethod
    def get_node(self, label, id_value, id_key="id"):
        """
        Get a node by label and identifier.

        Parameters:
        - label: Label of the node.
        - id_value: Unique identifier value of the node.
        - id_key: Property key used as the unique identifier.

        Returns:
        - The matching node.
        """
        pass

    @abstractmethod
    def delete_node(self, label, id_value, id_key="id"):
        """
        Delete a specified node.

        Parameters:
        - label: Label of the node.
        - id_value: Unique identifier value of the node.
        - id_key: Property key used as the unique identifier.
        """
        pass

    @abstractmethod
    def delete_nodes(self, label, id_values, id_key="id"):
        """
        Delete multiple nodes.

        Parameters:
        - label: Label of the nodes.
        - id_values: List of unique identifier values for the nodes.
        - id_key: Property key used as the unique identifier.
        """
        pass

    @abstractmethod
    def upsert_relationship(self, start_node_label, start_node_id_value,
                            end_node_label, end_node_id_value,
                            rel_type, properties, upsert_nodes=True,
                            start_node_id_key="id", end_node_id_key="id"):
        """
        Insert or update a relationship.

        Parameters:
        - start_node_label: Label of the start node.
        - start_node_id_value: Unique identifier value of the start node.
        - end_node_label: Label of the end node.
        - end_node_id_value: Unique identifier value of the end node.
        - rel_type: Type of the relationship.
        - properties: Properties of the relationship.
        - upsert_nodes: Whether to insert or update nodes.
        - start_node_id_key: Property key used as the unique identifier for the start node.
        - end_node_id_key: Property key used as the unique identifier for the end node.
        """
        pass

    @abstractmethod
    def upsert_relationships(self, start_node_label, end_node_label, rel_type,
                             relationships, upsert_nodes=True, start_node_id_key="id",
                             end_node_id_key="id"):
        """
        Insert or update multiple relationships.

        Parameters:
        - start_node_label: Label of the start node.
        - end_node_label: Label of the end node.
        - rel_type: Type of the relationship.
        - relationships: List of relationships.
        - upsert_nodes: Whether to insert or update nodes.
        - start_node_id_key: Property key used as the unique identifier for the start node.
        - end_node_id_key: Property key used as the unique identifier for the end node.
        """
        pass

    @abstractmethod
    def delete_relationship(self, start_node_label, start_node_id_value,
                            end_node_label, end_node_id_value,
                            rel_type, start_node_id_key="id", end_node_id_key="id"):
        """
        Delete a specified relationship.

        Parameters:
        - start_node_label: Label of the start node.
        - start_node_id_value: Unique identifier value of the start node.
        - end_node_label: Label of the end node.
        - end_node_id_value: Unique identifier value of the end node.
        - rel_type: Type of the relationship.
        - start_node_id_key: Property key used as the unique identifier for the start node.
        - end_node_id_key: Property key used as the unique identifier for the end node.
        """
        pass

    @abstractmethod
    def delete_relationships(self, start_node_label, start_node_id_values,
                             end_node_label, end_node_id_values, rel_type,
                             start_node_id_key="id", end_node_id_key="id"):
        """
        Delete multiple relationships.

        Parameters:
        - start_node_label: Label of the start node.
        - start_node_id_values: List of unique identifier values for the start nodes.
        - end_node_label: Label of the end node.
        - end_node_id_values: List of unique identifier values for the end nodes.
        - rel_type: Type of the relationship.
        - start_node_id_key: Property key used as the unique identifier for the start node.
        - end_node_id_key: Property key used as the unique identifier for the end node.
        """
        pass

    @abstractmethod
    def create_index(self, label, property_key, index_name=None):
        """
        Create a node index.

        Parameters:
        - label: Label of the node.
        - property_key: Property key used for indexing.
        - index_name: Name of the index (optional).
        """
        pass

    @abstractmethod
    def create_text_index(self, labels, property_keys, index_name=None):
        """
        Create a text index.

        Parameters:
        - labels: List of node labels.
        - property_keys: List of property keys used for indexing.
        - index_name: Name of the index (optional).
        """
        pass

    @abstractmethod
    def create_vector_index(self, label, property_key, index_name=None,
                            vector_dimensions=768, metric_type="cosine",
                            hnsw_m=None, hnsw_ef_construction=None):
        """
        Create a vector index.

        Parameters:
        - label: Label of the node.
        - property_key: Property key used for indexing.
        - index_name: Name of the index (optional).
        - vector_dimensions: Dimensionality of the vectors, default is 768.
        - metric_type: Type of distance measure, default is "cosine".
        - hnsw_m: m parameter of the HNSW algorithm, default to None (for m=16)
        - hnsw_ef_construction: ef_construction parameter of the HNSW algorithm, default to None (for ef_construction=100)
        """
        pass

    @abstractmethod
    def delete_index(self, index_name):
        """
        Delete a specified index.

        Parameters:
        - index_name: Name of the index.
        """
        pass

    @abstractmethod
    def text_search(self, query_string, label_constraints=None, topk=10, index_name=None):
        """
        Perform a text search.

        Parameters:
        - query_string: Query string.
        - label_constraints: Label constraints (optional).
        - topk: Number of top results to return, default is 10.
        - index_name: Name of the index (optional).

        Returns:
        - List of search results.
        """
        pass

    @abstractmethod
    def vector_search(self, label, property_key, query_text_or_vector, topk=10, index_name=None, ef_search=None):
        """
        Perform a vector search.

        Parameters:
        - label: Label of the node.
        - property_key: Property key used for indexing.
        - query_text_or_vector: Query text or vector.
        - topk: Number of top results to return, default is 10.
        - index_name: Name of the index (optional).
        - ef_search: ef_search parameter of the HNSW algorithm, specify number of potential candicates

        Returns:
        - List of search results.
        """
        pass

    @abstractmethod
    def execute_pagerank(self, iterations=20, damping_factor=0.85):
        """
        Execute the PageRank algorithm.

        Parameters:
        - iterations: Number of iterations, default is 20.
        - damping_factor: Damping factor, default is 0.85.
        """
        pass

    @abstractmethod
    def get_pagerank_scores(self, start_nodes, target_type):
        """
        Get PageRank scores.

        Parameters:
        - start_nodes: Start nodes.
        - target_type: Target node type.

        Returns:
        - PageRank scores.
        """
        pass

    @abstractmethod
    def run_script(self, script):
        """
        Execute a script.

        Parameters:
        - script: Script to be executed.
        """
        pass

    @abstractmethod
    def get_all_entity_labels(self):
        """
        Get all entity labels.

        Returns:
        - List of entity labels.
        """
        pass



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import json
from string import Template
from typing import List, Optional

from kag.common.base.prompt_op import PromptOp
from knext.schema.client import SchemaClient


class OpenIENERPrompt(PromptOp):

    template_en = """
    {
    "instruction": "You're a very effective entity extraction system. Please extract all the entities that are important for knowledge build and question, along with type, category and a brief description of the entity. The description of the entity is based on your OWN KNOWLEDGE AND UNDERSTANDING and does not need to be limited to the context. the entity's category belongs taxonomically to one of the items defined by schema, please also output the category. Note: Type refers to a specific, well-defined classification, such as Professor, Actor, while category is a broader group or class that may contain more than one type, such as Person, Works. Return an empty list if the entity type does not exist. Please respond in the format of a JSON string.You can refer to the example for extraction.",
    "schema": $schema,
    "example": [
        {
            "input": "The Rezort\nThe Rezort is a 2015 British zombie horror film directed by Steve Barker and written by Paul Gerstenberger.\n It stars Dougray Scott, Jessica De Gouw and Martin McCann.\n After humanity wins a devastating war against zombies, the few remaining undead are kept on a secure island, where they are hunted for sport.\n When something goes wrong with the island's security, the guests must face the possibility of a new outbreak.",
            "output": [
                        {
                            "entity": "The Rezort",
                            "type": "Movie",
                            "category": "Works",
                            "description": "A 2015 British zombie horror film directed by Steve Barker and written by Paul Gerstenberger."
                        },
                        {
                            "entity": "2015",
                            "type": "Year",
                            "category": "Date",
                            "description": "The year the movie 'The Rezort' was released."
                        },
                        {
                            "entity": "British",
                            "type": "Nationality",
                            "category": "GeographicLocation",
                            "description": "Great Britain, the island that includes England, Scotland, and Wales."
                        },
                        {
                            "entity": "Steve Barker",
                            "type": "Director",
                            "category": "Person",
                            "description": "Steve Barker is an English film director and screenwriter."
                        },
                        {
                            "entity": "Paul Gerstenberger",
                            "type": "Writer",
                            "category": "Person",
                            "description": "Paul is a writer and producer, known for The Rezort (2015), Primeval (2007) and House of Anubis (2011)."
                        },
                        {
                            "entity": "Dougray Scott",
                            "type": "Actor",
                            "category": "Person",
                            "description": "Stephen Dougray Scott (born 26 November 1965) is a Scottish actor."
                        },
                        {
                            "entity": "Jessica De Gouw",
                            "type": "Actor",
                            "category": "Person",
                            "description": "Jessica Elise De Gouw (born 15 February 1988) is an Australian actress. "
                        },
                        {
                            "entity": "Martin McCann",
                            "type": "Actor",
                            "category": "Person",
                            "description": "Martin McCann is an actor from Northern Ireland. In 2020, he was listed as number 48 on The Irish Times list of Ireland's greatest film actors"
                        }
                    ]
        }
    ],
    "input": "$input"
}    
        """

    template_zh = """
    {
        "instruction": "你是命名实体识别的专家。请从输入中提取与模式定义匹配的实体。如果不存在该类型的实体，请返回一个空列表。请以JSON字符串格式回应。你可以参照example进行抽取。",
        "schema": $schema,
        "example": [
            {
                "input": "《Rezort》\n《Rezort》是一部 2015 年英国僵尸恐怖片，由史蒂夫·巴克执导，保罗·格斯滕伯格编剧。\n 该片由道格瑞·斯科特、杰西卡·德·古维和马丁·麦凯恩主演。\n 在人类赢得与僵尸的毁灭性战争后，剩下的少数不死生物被关在一个安全的岛屿上，在那里他们被猎杀作为消遣。\n 当岛上的安全出现问题时，客人们必须面对新一轮疫情爆发的可能性。",
                "output": [
                            {
                                "entity": "The Rezort",
                                "type": "Movie",
                                "category": "Works",
                                "description": "一部 2015 年英国僵尸恐怖片，由史蒂夫·巴克执导，保罗·格斯滕伯格编剧。"
                            },
                            {
                                "entity": "2015",
                                "type": "Year",
                                "category": "Date",
                                "description": "电影《The Rezort》上映的年份。"
                            },
                            {
                                "entity": "英国",
                                "type": "Nationality",
                                "category": "GeographicLocation",
                                "description": "大不列颠，包括英格兰、苏格兰和威尔士的岛屿。"
                            },
                            {
                                "entity": "史蒂夫·巴克",
                                "type": "Director",
                                "category": "Person",
                                "description": "史蒂夫·巴克 是一名英国电影导演和剧作家"
                            },
                            {
                                "entity": "保罗·格斯滕伯格",
                                "type": "Writer",
                                "category": "Person",
                                "description": "保罗·格斯滕伯格 (Paul Gerstenberger) 是一名作家和制片人，因《The Rezort》（2015 年）、《Primeval》（2007 年）和《House of Anubis》（2011 年）而闻名。"
                            },
                            {
                                "entity": "道格雷·斯科特",
                                "type": "Actor",
                                "category": "Person",
                                "description": "斯蒂芬·道格雷·斯科特 (Stephen Dougray Scott，1965 年 11 月 26 日出生) 是一位苏格兰演员。"
                            },
                            {
                                "entity": "杰西卡·德·古维",
                                "type": "Actor",
                                "category": "Person",
                                "description": "杰西卡·伊莉斯·德·古维 (Jessica Elise De Gouw，1988 年 2 月 15 日出生) 是一位澳大利亚女演员。"
                            },
                            {
                                "entity": "马丁·麦肯",
                                "type": "Actor",
                                "category": "Person",
                                "description": "马丁·麦肯是来自北爱尔兰的演员。2020 年，他在《爱尔兰时报》爱尔兰最伟大电影演员名单中排名第 48 位"
                            }
                        ]
            }
        ],
        "input": "$input"
    }    
        """

    def __init__(
            self, language: Optional[str] = "en", **kwargs
    ):
        super().__init__(language, **kwargs)
        self.schema = SchemaClient(project_id=self.project_id).extract_types()
        self.template = Template(self.template).safe_substitute(schema=self.schema)

    @property
    def template_variables(self) -> List[str]:
        return ["input"]

    def parse_response(self, response: str, **kwargs):
        rsp = response
        if isinstance(rsp, str):
            rsp = json.loads(rsp)
        if isinstance(rsp, dict) and "output" in rsp:
            rsp = rsp["output"]
        if isinstance(rsp, dict) and "named_entities" in rsp:
            entities = rsp["named_entities"]
        else:
            entities = rsp

        return entities



import logging

from kag.interface.retriever.kg_retriever_abc import KGRetrieverABC
from kag.solver.logic.core_modules.common.base_model import LogicNode
from kag.solver.logic.core_modules.common.one_hop_graph import KgGraph
from kag.solver.logic.core_modules.common.schema_utils import SchemaUtils
from kag.solver.logic.core_modules.common.text_sim_by_vector import TextSimilarity
from kag.solver.logic.core_modules.op_executor.op_executor import OpExecutor
from kag.solver.logic.core_modules.op_executor.op_retrieval.module.get_spo_executor import GetSPOExecutor
from kag.solver.logic.core_modules.op_executor.op_retrieval.module.search_s import SearchS
from kag.solver.logic.core_modules.parser.logic_node_parser import GetSPONode
from kag.solver.logic.core_modules.retriver.entity_linker import EntityLinkerBase
from kag.solver.logic.core_modules.retriver.graph_retriver.dsl_executor import DslRunner

logger = logging.getLogger()


class RetrievalExecutor(OpExecutor):
    def __init__(self, nl_query: str, kg_graph: KgGraph, schema: SchemaUtils, retrieval_spo: KGRetrieverABC, el: EntityLinkerBase,
                 dsl_runner: DslRunner, debug_info: dict, text_similarity: TextSimilarity=None,**kwargs):
        super().__init__(nl_query, kg_graph, schema, debug_info, **kwargs)
        self.query_one_graph_cache = {}
        self.op_register_map = {
            'get_spo': GetSPOExecutor(nl_query, kg_graph, schema, retrieval_spo, el, dsl_runner, self.query_one_graph_cache, self.debug_info, text_similarity,KAG_PROJECT_ID = kwargs.get('KAG_PROJECT_ID')),
            'search_s': SearchS(nl_query, kg_graph, schema, self.debug_info,KAG_PROJECT_ID = kwargs.get('KAG_PROJECT_ID'))
        }

    def is_this_op(self, logic_node: LogicNode) -> bool:
        return isinstance(logic_node, GetSPONode)

    def executor(self, logic_node: LogicNode, req_id: str, param: dict) -> KgGraph:
        op = self.op_register_map.get(logic_node.operator, None)
        if op is None:
            return KgGraph()
        try:
            cur_kg_graph = op.executor(logic_node, req_id, param)
            if cur_kg_graph is not None:
                self.kg_graph.merge_kg_graph(cur_kg_graph)
        except Exception as e:
            logger.warning(f"op {logic_node.operator} run failed! {e}", exc_info=True)
        return self.kg_graph



import os
from abc import ABC, abstractmethod
from typing import List

from kag.solver.common.base import KagBaseModule
from kag.solver.logic.core_modules.common.base_model import LFPlanResult


class LFPlannerABC(KagBaseModule, ABC):
    """
    Initializes the base planner.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @abstractmethod
    def lf_planing(self, question, llm_output=None) -> List[LFPlanResult]:
        """
        Method that should be implemented by all subclasses for planning logic.
        This is a default impl

         :
        question (str): The question or task to plan.
        llm_output (Any, optional): Output from the LLM module. Defaults to None.

        Returns:
        list of LFPlanResult
        """
        pass


#
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import json
import logging

from kag.common.base.prompt_op import PromptOp

logger = logging.getLogger(__name__)




class AnalyzeTablePrompt(PromptOp):
    template_zh: str = """你是一个分析表格的专家, 从table中提取信息并分析，最后返回表格有效信息"""
    template_en: str = """You are an expert in knowledge graph extraction. Based on the schema defined by the constraint, extract all entities and their attributes from the input. Return NAN for attributes not explicitly mentioned in the input. Output the results in standard JSON format, as a list."""

    def __init__(
        self,
        language: str = "zh",
    ):
        super().__init__(
            language=language,
        )

    def build_prompt(self, variables) -> str:
        return json.dumps(
            {
                "instruction": self.template,
                "table": variables.get("table",""),
            },
            ensure_ascii=False,
        )

    def parse_response(self, response: str, **kwargs):
        return response




#
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import json
import logging
from abc import ABC
from typing import List, Dict

from kag.common.base.prompt_op import PromptOp
from knext.schema.client import SchemaClient
from knext.schema.model.base import BaseSpgType, SpgTypeEnum
from knext.schema.model.schema_helper import SPGTypeName
from kag.builder.model.spg_record import SPGRecord

logger = logging.getLogger(__name__)


class SPGPrompt(PromptOp, ABC):
    spg_types: Dict[str, BaseSpgType]
    ignored_types: List[str] = ["Chunk"]
    ignored_properties: List[str] = ["id", "name", "description", "stdId", "eventTime", "desc", "semanticType"]
    ignored_relations: List[str] = ["isA"]
    basic_types = {"Text": "文本", "Integer": "整型", "Float": "浮点型"}

    def __init__(
        self,
        spg_type_names: List[SPGTypeName],
        language: str = "zh",
        **kwargs,
    ):
        super().__init__(language=language, **kwargs)
        self.all_schema_types = SchemaClient(project_id=self.project_id).load()
        self.spg_type_names = spg_type_names
        if not spg_type_names:
            self.spg_types = self.all_schema_types
        else:
            self.spg_types = {k: v for k, v in self.all_schema_types.items() if k in spg_type_names}
        self.schema_list = []

        self._init_render_variables()

    @property
    def template_variables(self) -> List[str]:
        return ["schema", "input"]

    def _init_render_variables(self):
        self.type_en_to_zh = {"Text": "文本", "Integer": "整型", "Float": "浮点型"}
        self.type_zh_to_en = {
            "文本": "Text",
            "整型": "Integer",
            "浮点型": "Float",
        }
        self.prop_en_to_zh = {}
        self.prop_zh_to_en = {}
        for type_name, spg_type in self.all_schema_types.items():
            self.type_en_to_zh[type_name] = spg_type.name_zh
            self.type_en_to_zh[spg_type.name_zh] = type_name
            self.prop_zh_to_en[type_name] = {}
            self.prop_en_to_zh[type_name] = {}
            for _prop in spg_type.properties.values():
                if _prop.name in self.ignored_properties:
                    continue
                self.prop_en_to_zh[type_name][_prop.name] = _prop.name_zh
                self.prop_zh_to_en[type_name][_prop.name_zh] = _prop.name
            for _rel in spg_type.relations.values():
                if _rel.is_dynamic:
                    continue
                self.prop_en_to_zh[type_name][_rel.name] = _rel.name_zh
                self.prop_zh_to_en[type_name][_rel.name_zh] = _rel.name

    def _render(self):
        raise NotImplementedError


class SPG_KGPrompt(SPGPrompt):
    template_zh: str = """
    {
        "instruction": "你是一个图谱知识抽取的专家, 基于constraint 定义的schema，从input 中抽取出所有的实体及其属性，input中未明确提及的属性返回NAN，以标准json 格式输出，结果返回list",
        "schema": $schema,
        "example": [
        {
            "input": "甲状腺结节是指在甲状腺内的肿块，可随吞咽动作随甲状腺而上下移动，是临床常见的病症，可由多种病因引起。临床上有多种甲状腺疾病，如甲状腺退行性变、炎症、自身免疫以及新生物等都可以表现为结节。甲状腺结节可以单发，也可以多发，多发结节比单发结节的发病率高，但单发结节甲状腺癌的发生率较高。患者通常可以选择在普外科，甲状腺外科，内分泌科，头颈外科挂号就诊。有些患者可以触摸到自己颈部前方的结节。在大多情况下，甲状腺结节没有任何症状，甲状腺功能也是正常的。甲状腺结节进展为其它甲状腺疾病的概率只有1%。有些人会感觉到颈部疼痛、咽喉部异物感，或者存在压迫感。当甲状腺结节发生囊内自发性出血时，疼痛感会更加强烈。治疗方面，一般情况下可以用放射性碘治疗，复方碘口服液(Lugol液)等，或者服用抗甲状腺药物来抑制甲状腺激素的分泌。目前常用的抗甲状腺药物是硫脲类化合物，包括硫氧嘧啶类的丙基硫氧嘧啶(PTU)和甲基硫氧嘧啶(MTU)及咪唑类的甲硫咪唑和卡比马唑。",
            "schema": {
                "Disease": {
                    "properties": {
                        "complication": "并发症",
                        "commonSymptom": "常见症状",
                        "applicableMedicine": "适用药品",
                        "department": "就诊科室",
                        "diseaseSite": "发病部位",
                    }
                },"Medicine": {
                    "properties": {
                    }
                }
            }
            "output": [
                {
                    "entity": "甲状腺结节",
                    "category":"Disease"
                    "properties": {
                        "complication": "甲状腺癌",
                        "commonSymptom": ["颈部疼痛", "咽喉部异物感", "压迫感"],
                        "applicableMedicine": ["复方碘口服液(Lugol液)", "丙基硫氧嘧啶(PTU)", "甲基硫氧嘧啶(MTU)", "甲硫咪唑", "卡比马唑"],
                        "department": ["普外科", "甲状腺外科", "内分泌科", "头颈外科"],
                        "diseaseSite": "甲状腺",
                    }
                },{
                    "entity":"复方碘口服液(Lugol液)",
                    "category":"Medicine"
                },{
                    "entity":"丙基硫氧嘧啶(PTU)",
                    "category":"Medicine"
                },{
                    "entity":"甲基硫氧嘧啶(MTU)",
                    "category":"Medicine"
                },{
                    "entity":"甲硫咪唑",
                    "category":"Medicine"
                },{
                    "entity":"卡比马唑",
                    "category":"Medicine"
                }
            ],
    "input": "$input"
    }
    """

    template_en: str = template_zh

    def __init__(
        self,
        spg_type_names: List[SPGTypeName],
        language: str = "zh",
        **kwargs
    ):
        super().__init__(
            spg_type_names=spg_type_names,
            language=language,
            **kwargs
        )
        self._render()

    def build_prompt(self, variables: Dict[str, str]) -> str:
        schema = {}
        for tmpSchema in self.schema_list:
            schema.update(tmpSchema)

        return super().build_prompt({"schema": schema, "input": variables.get("input")})

    def parse_response(self, response: str, **kwargs) -> List[SPGRecord]:
        rsp = response
        if isinstance(rsp, str):
            rsp = json.loads(rsp)
        if isinstance(rsp, dict) and "output" in rsp:
            rsp = rsp["output"]
        if isinstance(rsp, dict) and "named_entities" in rsp:
            entities = rsp["named_entities"]
        else:
            entities = rsp

        return entities

    def _render(self):
        spo_list = []
        for type_name, spg_type in self.spg_types.items():
            if spg_type.spg_type_enum not in [SpgTypeEnum.Entity, SpgTypeEnum.Concept, SpgTypeEnum.Event]:
                continue
            constraint = {}
            properties = {}
            properties.update(
                {
                    v.name: (f"{v.name_zh}" if not v.desc else f"{v.name_zh}，{v.desc}") if self.language == "zh" else (f"{v.name}" if not v.desc else f"{v.name}, {v.desc}")
                    for k, v in spg_type.properties.items()
                    if k not in self.ignored_properties
                }
            )
            properties.update(
                {
                    f"{v.name}#{v.object_type_name_en}": (
                        f"{v.name_zh}，类型是{v.object_type_name_zh}"
                        if not v.desc
                        else f"{v.name_zh}，{v.desc}，类型是{v.object_type_name_zh}"
                    ) if self.language == "zh" else (
                        f"{v.name}, the type is {v.object_type_name_en}"
                        if not v.desc
                        else f"{v.name}，{v.desc}, the type is {v.object_type_name_en}"
                    )
                    for k, v in spg_type.relations.items()
                    if not v.is_dynamic and k not in self.ignored_relations
                }
            )
            constraint.update({"properties": properties})
            spo_list.append({type_name: constraint})

        self.schema_list = spo_list



from typing import Union

from kag.solver.logic.core_modules.common.base_model import LogicNode
from kag.solver.logic.core_modules.common.one_hop_graph import KgGraph
from kag.solver.logic.core_modules.common.schema_utils import SchemaUtils
from kag.solver.logic.core_modules.op_executor.op_deduce.module.choice import ChoiceOp
from kag.solver.logic.core_modules.op_executor.op_deduce.module.entailment import EntailmentOp
from kag.solver.logic.core_modules.op_executor.op_deduce.module.judgement import JudgementOp
from kag.solver.logic.core_modules.op_executor.op_deduce.module.multi_choice import MultiChoiceOp
from kag.solver.logic.core_modules.op_executor.op_executor import OpExecutor
from kag.solver.logic.core_modules.parser.logic_node_parser import FilterNode, VerifyNode, \
    ExtractorNode, DeduceNode
from kag.solver.logic.core_modules.rule_runner.rule_runner import OpRunner


class DeduceExecutor(OpExecutor):
    def __init__(self, nl_query: str, kg_graph: KgGraph, schema: SchemaUtils, rule_runner: OpRunner, debug_info: dict,
                 **kwargs):
        super().__init__(nl_query, kg_graph, schema, debug_info, **kwargs)
        self.KAG_PROJECT_ID = kwargs.get('KAG_PROJECT_ID')
        self.rule_runner = rule_runner
        self.op_register_map = {
            'verify': self.rule_runner.run_verify_op,
            'filter': self.rule_runner.run_filter_op,
            'extractor': self.rule_runner.run_extractor_op
        }

    def _deduce_call(self, node: DeduceNode, req_id: str, param: dict) -> list:
        op_mapping = {
            'choice': ChoiceOp(self.nl_query, self.kg_graph, self.schema, self.debug_info,KAG_PROJECT_ID = self.KAG_PROJECT_ID),
            'multiChoice': MultiChoiceOp(self.nl_query, self.kg_graph, self.schema, self.debug_info,KAG_PROJECT_ID = self.KAG_PROJECT_ID),
            'entailment': EntailmentOp(self.nl_query, self.kg_graph, self.schema, self.debug_info,KAG_PROJECT_ID = self.KAG_PROJECT_ID),
            'judgement': JudgementOp(self.nl_query, self.kg_graph, self.schema, self.debug_info,KAG_PROJECT_ID = self.KAG_PROJECT_ID)
        }
        result = []
        for op in node.deduce_ops:
            if_answered, answer = op_mapping[op].executor(node, req_id, param)
            if if_answered:
                result.append(answer)
        return result

    def is_this_op(self, logic_node: LogicNode) -> bool:
        return isinstance(logic_node, (DeduceNode, FilterNode, VerifyNode, ExtractorNode))

    def executor(self, logic_node: LogicNode, req_id: str, param: dict) -> Union[KgGraph, list]:
        if isinstance(logic_node, DeduceNode):
            return self._deduce_call(logic_node, req_id, param)
        op_func = self.op_register_map.get(logic_node.operator, None)
        if op_func is None:
            return self.kg_graph
        return op_func(logic_node)



# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import io
import os
import threading
import tarfile
import requests
from typing import Any, Union, Iterable, Dict
from kag.common.vectorizer.vectorizer import Vectorizer


EmbeddingVector = Iterable[float]


class LocalBGEVectorizer(Vectorizer):
    """
    Invoke local bge embedding models to turn texts into embedding vectors.
    """

    _local_model_map = {}
    _lock = threading.Lock()

    def __init__(self, config: Dict[str, Any]):
        super().__init__(config)
        path = config.get("path")
        if path is None:
            message = "model path is required"
            raise RuntimeError(message)
        url = config.get("url")
        path = os.path.expanduser(path)
        config_path = os.path.join(path, "config.json")
        if not os.path.isfile(config_path):
            if url is None:
                message = f"model not found at {path!r}, nor model url specified"
                raise RuntimeError(message)
            self._download_model(path, url)
        default_chinese_query_instruction_for_retrieval = "为这个句子生成表示以用于向量检索："
        default_english_query_instruction_for_retrieval = "Represent this sentence for searching relevant passages:"
        if "BAAI/bge-base-zh-v1.5" in path:
            default_query_instruction_for_retrieval = default_chinese_query_instruction_for_retrieval
        else:
            default_query_instruction_for_retrieval = default_english_query_instruction_for_retrieval
        query_instruction_for_retrieval = config.get("query_instruction_for_retrieval", default_query_instruction_for_retrieval)
        self._path = path
        self._url = url
        self._query_instruction_for_retrieval = query_instruction_for_retrieval
        with self._lock:
            if path in self._local_model_map:
                self._model = self._local_model_map[path]
            else:
                self._model = self._load_model(path)
                self._local_model_map[path] = self._model

    @classmethod
    def _from_config(cls, config: Dict[str, Any]) -> Vectorizer:
        """
        Create vectorizer from `config`.

        :param config: vectorizer config
        :type config: Dict[str, Any]
        :return: vectorizer instance
        :rtype: Vectorizer
        """
        vectorizer = cls(config)
        return vectorizer

    def _download_model(self, path, url):
        res = requests.get(url)
        with io.BytesIO(res.content) as fileobj:
            with tarfile.open(fileobj=fileobj) as tar:
                tar.extractall(path=path)
        config_path = os.path.join(path, "config.json")
        if not os.path.isfile(config_path):
            message = f"model config not found at {config_path!r}, url {url!r} specified an invalid model"
            raise RuntimeError(message)

    def _load_model(self, path):
        # We need to import sklearn at first, otherwise sklearn will fail on macOS with m chip.
        import sklearn
        from FlagEmbedding import FlagModel

        print(f"Loading FlagModel from {path!r} with query_instruction_for_retrieval={self._query_instruction_for_retrieval!r}")
        model = FlagModel(path,
                          query_instruction_for_retrieval=self._query_instruction_for_retrieval,
                          use_fp16=True)
        return model

    def vectorize(self, texts: Union[str, Iterable[str]]) -> Union[EmbeddingVector, Iterable[EmbeddingVector]]:
        """
        Vectorize a text string into an embedding vector or multiple text strings into
        multiple embedding vectors.

        :param texts: texts to vectorize
        :type texts: str or Iterable[str]
        :return: embedding vectors of the texts
        :rtype: EmbeddingVector or Iterable[EmbeddingVector]
        """
        result = self._model.encode(texts)
        return result.tolist()



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

from kag.builder.component.reader.csv_reader import CSVReader
from kag.builder.component.reader.pdf_reader import PDFReader
from kag.builder.component.reader.json_reader import JSONReader
from kag.builder.component.reader.markdown_reader import MarkDownReader
from kag.builder.component.reader.docx_reader import DocxReader
from kag.builder.component.reader.txt_reader import TXTReader
from kag.builder.component.reader.dataset_reader import HotpotqaCorpusReader, TwowikiCorpusReader, MusiqueCorpusReader
from kag.builder.component.reader.yuque_reader import YuqueReader

__all__ = [
    "TXTReader",
    "PDFReader",
    "MarkDownReader",
    "JSONReader",
    "HotpotqaCorpusReader",
    "MusiqueCorpusReader",
    "TwowikiCorpusReader",
    "YuqueReader",
    "CSVReader",
    "DocxReader",
]



import itertools
import json
import re
from typing import List


class Identifer:
    def __init__(self, alias_name):
        self.alias_name = alias_name

    def __repr__(self):
        return self.alias_name

    def __str__(self):
        return self.alias_name

    def __eq__(self, other):
        if isinstance(other, Identifer):
            return self.alias_name == other.alias_name
        if isinstance(other, str):
            return self.alias_name == other
        return False

    def __hash__(self):
        return hash(self.alias_name)


class TypeInfo:
    def __init__(self, entity_type=None, entity_type_zh=None):
        self.entity_type = entity_type
        self.entity_type_zh = entity_type_zh

    def __repr__(self):
        return f"en:{self.entity_type} zh:{self.entity_type_zh}"


def parse_entity(raw_entity):
    if raw_entity is None:
        return []
    entity_parts = re.findall(r'(?:`(.+?)`|([^|]+))', raw_entity)
    return [part.replace('``', '|') if part else escaping_part for escaping_part, part in entity_parts]


class SPOBase:
    def __init__(self):
        self.alias_name: Identifer = None
        self.type_set: List[TypeInfo] = []
        self.is_attribute = False
        self.value_list = []

    def __repr__(self):
        return f"{self.alias_name}:{self.get_entity_first_type_or_en()}"

    def get_value_list_str(self):
        return [f"{self.alias_name}.{k}={v}" for k,v in self.value_list]

    def get_mention_name(self):
        return ""

    def get_type_with_gql_format(self):
        entity_types = self.get_entity_type_set()
        entity_zh_types = self.get_entity_type_zh_set()
        if len(entity_types) == 0 and len(entity_zh_types) == 0:
            return None
        if None in entity_types and None in entity_zh_types:
            raise RuntimeError(f"None type in entity type en {entity_types} zh {entity_zh_types}")
        if len(entity_types) > 0:
            return "|".join(entity_types)
        if len(entity_zh_types) > 0:
            return "|".join(entity_zh_types)

    def get_entity_first_type(self):
        type_list = list(self.get_entity_type_set())
        if len(type_list) == 0:
            return None
        return type_list[0]

    def get_entity_first_type_or_en(self):
        en = list(self.get_entity_type_set())
        zh = list(self.get_entity_type_zh_set())
        if len(zh) > 0:
            return zh[0]
        elif len(en) > 0:
            return en[0]
        else:
            return None

    def get_entity_type_or_zh_list(self):
        ret = []
        for entity_type_info in self.type_set:
            if entity_type_info.entity_type is not None:
                ret.append(entity_type_info.entity_type)
            elif entity_type_info.entity_type_zh is not None:
                ret.append(entity_type_info.entity_type_zh)
        return ret

    def get_entity_first_type_or_zh(self):
        en = list(self.get_entity_type_set())
        zh = list(self.get_entity_type_zh_set())
        if len(en) > 0:
            return en[0]
        elif len(zh) > 0:
            return zh[0]
        else:
            return None

    def get_entity_type_set(self):
        entity_types = []
        for entity_type_info in self.type_set:
            if entity_type_info.entity_type is not None:
                entity_types.append(entity_type_info.entity_type)
        return set(entity_types)

    def get_entity_type_zh_set(self):
        entity_types = []
        for entity_type_info in self.type_set:
            if entity_type_info.entity_type_zh is not None:
                entity_types.append(entity_type_info.entity_type_zh)
        return set(entity_types)


class SPORelation(SPOBase):
    def __init__(self, alias_name=None, rel_type=None, rel_type_zh=None):
        super().__init__()
        if rel_type is not None or rel_type_zh is not None:
            type_info = TypeInfo()
            type_info.entity_type = rel_type
            type_info.entity_type_zh = rel_type_zh
            self.type_set.append(type_info)
        self.alias_name: Identifer = None
        if alias_name is not None:
            self.alias_name = Identifer(alias_name)

        self.s: SPOBase = None
        self.o: SPOEntity = None

    def __str__(self):
        show = [f"{self.alias_name}:{self.get_entity_first_type_or_en()}"]
        show = show + self.get_value_list_str()
        return ",".join(show)

    @staticmethod
    def parse_logic_form(input_str):
        """
        Parses the logic form from the given input string and constructs a relation object.

        Parameters:
            input_str (str): The input string containing alias and entity types separated by ':'.

        Returns:
            SPORelation: A relation object with alias name and associated type set.
        """

        rel_type_set = []

        # Split the input string into alias and entity_type_set parts
        split_input = input_str.split(':', 1)
        alias = split_input[0]
        # If entity_type_set exists, process it further
        if len(split_input) > 1:
            entity_type_part = split_input[1]

            entity_types = parse_entity(entity_type_part)
            for entity_type in entity_types:
                entity_type_obj = TypeInfo()
                entity_type_obj.entity_type_zh = entity_type
                rel_type_set.append(entity_type_obj)

        rel = SPORelation()
        rel.alias_name = Identifer(alias)
        rel.type_set = rel_type_set
        return rel


class SPOEntity(SPOBase):
    def __init__(self, entity_id=None, entity_type=None, entity_type_zh=None, entity_name=None, alias_name=None,
                 is_attribute=False):
        super().__init__()
        self.is_attribute = is_attribute
        self.id_set = []
        self.entity_name = entity_name
        self.alias_name: Identifer = None
        if alias_name is not None:
            self.alias_name = Identifer(alias_name)
        if entity_id is not None:
            self.id_set.append(entity_id)
        if entity_type is not None or entity_type_zh is not None:
            type_info = TypeInfo()
            type_info.entity_type = entity_type
            type_info.entity_type_zh = entity_type_zh
            self.type_set.append(type_info)

    def __str__(self):
        show = [f"{self.alias_name}:{self.get_entity_first_type_or_en()}{'' if self.entity_name is None else '[' + self.entity_name + ']'} "]
        show = show + self.get_value_list_str()
        return ",".join(show)

    def get_mention_name(self):
        return self.entity_name
    def generate_id_key(self):
        if len(self.id_set) == 0:
            return None
        id_str_set = ['"' + id_str + '"' for id_str in self.id_set]
        return ",".join(id_str_set)

    def generate_start_infos(self, prefix=None):
        if len(self.id_set) == 0:
            return []
        if len(self.type_set) == 0:
            return []

        id_type_info = list(itertools.product(self.id_set, self.type_set))
        return [{
            "alias": self.alias_name.alias_name,
            "id": info[0],
            "type": info[1].entity_type if '.' in info[1].entity_type else (
                                                                               prefix + '.' if prefix is not None else '') +
                                                                           info[1].entity_type
        } for info in id_type_info]

    @staticmethod
    def parse_logic_form(input_str):
        # # 正则表达式解析输入字符串
        match = re.match(r"([^:]+):?([^\[]+)?(\[[^\[]*\])?(\[[^\[]*\])?", input_str)
        if not match:
            return None

        # 提取和解构匹配的组件
        alias = match.group(1)
        entity_type_raw = match.group(2)
        entity_name_raw = match.group(3)
        entity_id_raw = match.group(4)

        # 处理entity_type_set
        entity_type_set = parse_entity(entity_type_raw)

        # 解析entity_name和entity_id_set
        entity_name = entity_name_raw.strip('][') if entity_name_raw else None
        entity_name = entity_name.strip('`') if entity_name else None
        entity_id_set = parse_entity(entity_id_raw.strip('][')) if entity_id_raw else []

        spo_entity = SPOEntity()
        spo_entity.id_set = entity_id_set
        spo_entity.alias_name = Identifer(alias)
        spo_entity.entity_name = entity_name
        for entity_type in entity_type_set:
            entity_type_obj = TypeInfo()
            entity_type_obj.entity_type_zh = entity_type
            entity_type_obj.entity_type = entity_type
            spo_entity.type_set.append(entity_type_obj)
        return spo_entity


class Entity:
    def __init__(self, entity_id=None, entity_type=None, entity_type_zh=None, entity_name=None, alias_name=None):
        self.id = entity_id
        self.type = entity_type
        self.entity_type_zh = entity_type_zh
        self.entity_name = entity_name
        self.alias_name = alias_name

    def __repr__(self):
        return f"{[self.entity_name, self.alias_name]}:{self.id}({self.type, self.entity_type_zh})"

    def save_args(self, id=None, type=None, entity_type_zh=None, entity_name=None, alias_name=None):
        self.id = id if id else self.id
        self.type = type if type else self.type
        self.entity_type_zh = entity_type_zh if entity_type_zh else self.entity_type_zh
        self.entity_name = entity_name if entity_name else self.entity_name
        self.alias_name = alias_name if alias_name else self.alias_name

    @staticmethod
    def parse_zh(entity_str):
        alias, type_zh, name = '', '', ''
        entity_str = entity_str.replace('：', ':')
        match_alias_type_entity = re.match(r'(.*):(.*)\[(.*)\]', entity_str)
        if match_alias_type_entity:
            alias, type_zh, name = match_alias_type_entity.groups()
        else:
            match_alias_type = re.match(r'(.*):(.*)', entity_str)
            if match_alias_type:
                alias, type_zh = match_alias_type.groups()
            else:
                alias = entity_str
        return Entity(entity_type_zh=type_zh.strip(), entity_name=name.strip(), alias_name=alias.strip())


class LogicNode:
    def __init__(self, operator, args):
        self.operator = operator
        self.args = args
        self.sub_query = args.get('sub_query', '')

    def __repr__(self):
        params = [f"{k}={v}" for k, v in self.args.items()]
        params_str = ','.join(params)
        return f"{self.operator}({params_str})"

    def to_dict(self):
        return json.loads(self.to_json())

    def to_json(self):
        return json.dumps(obj=self,
                          default=lambda x: x.__dict__, sort_keys=False, indent=2)

    def to_dsl(self):
        raise NotImplementedError("Subclasses should implement this method.")

    def to_std(self, args):
        for key, value in args.items():
            self.args[key] = value
        self.sub_query = args.get('sub_query', '')


class LFPlanResult:
    def __init__(self, query: str, lf_nodes: List[LogicNode]):
        self.query: str = query
        self.lf_nodes: List[LogicNode] = lf_nodes


# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import json
from typing import Optional, List

from kag.common.base.prompt_op import PromptOp


class OutlinePrompt(PromptOp):
    template_zh = """
{
    "instruction": "\n请理解input字段中的文本内容，识别文本的结构和组成部分，并帮我提取出以下内容的标题，可能有多个标题分散在文本的各个地方，仅返属于原文的回标题文本即可，不要返回其他任何内容，须按照python list的格式回答，具体形式请遵从example字段中给出的若干例子。",
    "input": "$input",
    "example": [
        {
            "input": "第8条 原 则

1.各成员方在制订或修正其法律和规章时，可采取必要措施以保护公众健康和营养，并促进对其社会经济和技术发展至关重要部门的公众利益，只要该措施符合本协议规定。

2.可能需要采取与本协议的规定相一致的适当的措施，以防止知识产权所有者滥用知识产权或藉以对贸易进行不合理限制或实行对国际间的技术转让产生不利影响的作法。

第二部分 关于知识产权的效力、范围及使用的标准

第1节 版权及相关权利

第9条 与《伯尔尼公约》的关系",
            "output": [
                "第8条 原 则",
                "第二部分 关于知识产权的效力、范围及使用的标准",
                "第1节 版权及相关权利",
                "第9条 与《伯尔尼公约》的关系"
            ],
        },
        {
            "input": "第16条 授予权利

1.已注册商标所有者应拥有阻止所有未经其同意的第三方在贸易中使用与已注册商标相同或相似的商品或服务的，其使用有可能招致混淆的相同或相似的标志。在对相同商品或服务使用相同标志的情况下，应推定存在混淆之可能。上述权利不应妨碍任何现行的优先权，也不应影响各成员方以使用为条件获得注册权的可能性。

2.1967《巴黎公约》第6条副则经对细节作必要修改后应适用于服务。在确定一个商标是否为知名商标时，各成员方应考虑到有关部分的公众对该商标的了解，包括由于该商标的推行而在有关成员方得到的了解。

3.1967《巴黎公约》第6条副则经对细节作必要修改后应适用于与已注册商标的商品和服务不相似的商品或服务，条件是该商标与该商品和服务有关的使用会表明该商品或服务与已注册商标所有者之间的联系，而且已注册商标所有者的利益有可能为此种使用所破坏。

第17条 例 外\n ",
            "output": [
                "第16条 授予权利",
                "第17条 例 外"
            ],
        },
        {
            "input":"的做法。

（4）此类使用应是非独占性的。

（5）此类使用应是不可转让的，除非是同享有此类使用的那部分企业或信誉一道转让。

（6）任何此类使用之授权，均应主要是为授权此类使用的成员方国内市场供应之目的。

（7）在被授权人的合法利益受到充分保护的条件下，当导致此类使用授权的情况下不复存在和可能不再产生时，有义务将其终止；应有动机的请求，主管当局应有权对上述情况的继续存在进行检查。

（8）考虑到授权的经济价值，应视具体情况向权利人支付充分的补偿金。

（9）任何与此类使用之授权有关的决定，其法律效力应接受该成员方境内更高当局的司法审查或其他独立审查。

（10）任何与为此类使用而提供的补偿金有关的决定，应接受成员方境内更高当局的司法审查或其他独立审查。
",
            "output": [],
        },
    ]
}    
    """

    template_en = """
{
    "instruction": "\nUnderstand the text content in the input field, identify the structure and components of the text, and help me extract the titles from the following content. There may be multiple titles scattered throughout the text. Only return the title texts that belong to the original text, and do not return any other content. The response must be in the format of a Python list, and the specific form should follow the examples given in the example field.",
    "input": "$input",
    "example": [
        {
            "input": "Article 8 Principles

1. In formulating or amending their laws and regulations, Members may take necessary measures to protect public health and nutrition, and to promote the public interest in sectors of vital importance to their socio-economic and technological development, provided that such measures are consistent with the provisions of this Agreement.

2. Appropriate measures may be needed to prevent the abuse of intellectual property rights by owners, or to prevent practices that restrict trade unjustifiably or adversely affect the international transfer of technology, in conformity with the provisions of this Agreement.

Part Two: Standards Concerning the Availability, Scope and Use of Intellectual Property Rights

Section 1 Copyright and Related Rights

Article 9 Relationship with the Berne Convention
",
            "output": [
                "Article 8 Principles",
                "Part Two: Standards Concerning the Availability, Scope and Use of Intellectual Property Rights",
                "Section 1 Copyright and Related Rights",
                "Article 9 Relationship with the Berne Convention"
            ],
        },
        {
            "input": "Article 16 Grant of Rights

1. Owners of registered trademarks shall have the right to prevent all third parties from using, without their consent, in the course of trade, any identical or similar signs for goods or services that are identical or similar to those for which the trademark is registered, where such use is likely to cause confusion. In the case of identical signs being used for identical goods or services, a likelihood of confusion shall be presumed. The above rights shall not prejudice any existing priority rights, nor shall they affect the possibility for Members to obtain registration rights conditional upon use.

2. The provisions of Article 6bis of the Paris Convention of 1967 shall apply to services with necessary modifications to the details. In determining whether a trademark is well-known, Members shall take into account the knowledge of the relevant public about that trademark, including the knowledge acquired in the relevant Member due to the promotion of that trademark.

3. The provisions of Article 6bis of the Paris Convention of 1967 shall apply to goods or services that are not similar to those for which the registered trademark is granted, provided that the use of the trademark in relation to those goods or services indicates a connection between the goods or services and the owner of the registered trademark, and that the interests of the owner of the registered trademark are likely to be harmed by such use.

Article 17 Exceptions
",
            "output": [
                "Article 16 Grant of Rights",
                "Article 17 Exceptions"
            ],
        },
        {
            "input": "by doing so.

(4) The use of this category should be non-exclusive.

(5) The use of this category should be non-transferable, unless it is transferred together with the part of the enterprise or reputation that enjoys the use of this category.

(6) Any authorization for such use should primarily be for the purpose of domestic market supply by the member party authorizing such use.

(7) There is an obligation to terminate it when the circumstances leading to the authorization for such use no longer exist and are unlikely to reoccur; upon motivated request, the competent authorities should have the right to examine the continued existence of the above circumstances.

(8) Adequate compensation should be paid to the right holder, taking into account the economic value of the authorization.

(9) Any decisions related to the authorization for such use should be subject to judicial review or other independent review by a higher authority within the territory of the member party.

(10) Any decisions related to the compensation provided for such use should be subject to judicial review or other independent review by a higher authority within the territory of the member party.",
            "output": [],    
        },
    ]
}    
    """

    def __init__(self, language: Optional[str] = "zh"):
        super().__init__(language)

    @property
    def template_variables(self) -> List[str]:
        return ["input"]

    def parse_response(self, response: str, **kwargs):
        if isinstance(response, str):
            response = json.loads(response)
        if isinstance(response, dict) and "output" in response:
            response = response["output"]

        outline = kwargs.get("outline", [])
        for r in response:
            outline.append(r)

        return outline



# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import os
from tenacity import retry, stop_after_attempt

from kag.common.base.prompt_op import PromptOp
from kag.common.vectorizer import Vectorizer
from knext.graph_algo.client import GraphAlgoClient
from kag.interface.retriever.chunk_retriever_abc import ChunkRetrieverABC
from typing import List, Dict

import numpy as np
import logging

from knext.reasoner.client import ReasonerClient
from knext.schema.client import CHUNK_TYPE, OTHER_TYPE
from knext.project.client import ProjectClient
from kag.common.utils import processing_phrases
from knext.search.client import SearchClient
from kag.solver.logic.core_modules.common.schema_utils import SchemaUtils
from kag.solver.logic.core_modules.config import LogicFormConfiguration

logger = logging.getLogger(__name__)


class DefaultRetriever(ChunkRetrieverABC):
    """
    KAGRetriever class for retrieving and processing knowledge graph data from a graph database.

    this retriever references the implementation of Hippoag for the combination of dpr & ppr, developer can define your Retriever

    Parameters:
    - project_id (str, optional): Project ID to load specific project configurations.
    - host_addr (str, optional): host addr to load specific server addr configurations.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.schema_util = SchemaUtils(LogicFormConfiguration(kwargs))

        self._init_search()

        self.ner_prompt = PromptOp.load(self.biz_scene, "question_ner")(language=self.language, project_id=self.project_id)
        self.std_prompt = PromptOp.load(self.biz_scene, "std")(language=self.language)

        self.pagerank_threshold = 0.9
        self.match_threshold = 0.8
        self.pagerank_weight = 0.5

        self.reranker_model_path = os.getenv("KAG_RETRIEVER_RERANKER_MODEL_PATH")
        if self.reranker_model_path:
            from kag.common.reranker.reranker import BGEReranker
            self.reranker = BGEReranker(self.reranker_model_path, use_fp16=True)
        else:
            self.reranker = None

        self.with_semantic = True

    def _init_search(self):
        self.sc: SearchClient = SearchClient(self.host_addr, self.project_id)
        vectorizer_config = eval(os.getenv("KAG_VECTORIZER", "{}"))
        if self.host_addr and self.project_id:
            config = ProjectClient(host_addr=self.host_addr, project_id=self.project_id).get_config(self.project_id)
            vectorizer_config.update(config.get("vectorizer", {}))

        self.vectorizer = Vectorizer.from_config(
            vectorizer_config
        )
        self.reason: ReasonerClient = ReasonerClient(self.host_addr, self.project_id)
        self.graph_algo = GraphAlgoClient(self.host_addr, self.project_id)




    @retry(stop=stop_after_attempt(3))
    def named_entity_recognition(self, query: str):
        """
        Perform named entity recognition.

        This method invokes the pre-configured service client (self.llm) to process the input query,
        using the named entity recognition (NER) prompt (self.ner_prompt).

        Parameters:
        query (str): The text input provided by the user or system for named entity recognition.

        Returns:
        The result returned by the service client, with the type and format depending on the used service.
        """
        return self.llm_module.invoke({"input": query}, self.ner_prompt)

    @retry(stop=stop_after_attempt(3))
    def named_entity_standardization(self, query: str, entities: List[Dict]):
        """
        Entity standardization function.

        This function calls a remote service to process the input query and named entities,
        standardizing the entities. This is useful for unifying different representations of the same entity in text,
        improving the performance of natural language processing tasks.

        Parameters:
        - query: A string containing the query with named entities.
        - entities: A list of dictionaries, each containing information about named entities.

        Returns:
        - The result of the remote service call, typically standardized named entity information.
        """
        return self.llm_module.invoke(
            {"input": query, "named_entities": entities}, self.std_prompt
        )

    @staticmethod
    def append_official_name(source_entities: List[Dict], entities_with_official_name: List[Dict]):
        """
        Appends official names to entities.

        Parameters:
        source_entities (List[Dict]): A list of source entities.
        entities_with_official_name (List[Dict]): A list of entities with official names.

        """
        tmp_dict = {}
        for tmp_entity in entities_with_official_name:
            name = tmp_entity["entity"]
            category = tmp_entity["category"]
            official_name = tmp_entity["official_name"]
            key = f"{category}{name}"
            tmp_dict[key] = official_name

        for tmp_entity in source_entities:
            name = tmp_entity["entity"]
            category = tmp_entity["category"]
            key = f"{category}{name}"
            if key in tmp_dict:
                official_name = tmp_dict[key]
                tmp_entity["official_name"] = official_name

    def calculate_sim_scores(self, query: str, doc_nums: int):
        """
        Calculate the vector similarity scores between a query and document chunks.

        Parameters:
        query (str): The user's query text.
        doc_nums (int): The number of document chunks to return.

        Returns:
        dict: A dictionary with keys as document chunk IDs and values as the vector similarity scores.
        """
        scores = dict()
        try:
            query_vector = self.vectorizer.vectorize(query)
            top_k = self.sc.search_vector(
                label=self.schema_util.get_label_within_prefix(CHUNK_TYPE),
                property_key="content",
                query_vector=query_vector,
                topk=doc_nums
            )
            scores = {item["node"]["id"]: item["score"] for item in top_k}
        except Exception as e:
            logger.error(
                f"run calculate_sim_scores failed, info: {e}", exc_info=True
            )
        return scores

    def calculate_pagerank_scores(self, start_nodes: List[Dict]):
        """
        Calculate and retrieve PageRank scores for the given starting nodes.

        Parameters:
        start_nodes (list): A list containing document fragment IDs to be used as starting nodes for the PageRank algorithm.

        Returns:
        ppr_doc_scores (dict): A dictionary containing each document fragment ID and its corresponding PageRank score.

        This method uses the PageRank algorithm in the graph store to compute scores for document fragments. If `start_nodes` is empty,
        it returns an empty dictionary. Otherwise, it attempts to retrieve PageRank scores from the graph store and converts the result
        into a dictionary format where keys are document fragment IDs and values are their respective PageRank scores. Any exceptions,
        such as failures in running `run_pagerank_igraph_chunk`, are logged.
        """
        scores = dict()
        if len(start_nodes) != 0:
            try:
                scores = self.graph_algo.calculate_pagerank_scores(
                    self.schema_util.get_label_within_prefix(CHUNK_TYPE),
                    start_nodes
                )
            except Exception as e:
                logger.error(
                    f"run calculate_pagerank_scores failed, info: {e}, start_nodes: {start_nodes}", exc_info=True
                )
        return scores

    def match_entities(self, queries: Dict[str, str], top_k: int = 1):
        """
        Match entities based on the provided queries.

        :param queries: A dictionary containing keywords and their labels.
        :param top_k: The number of top results to return. Default is 1.
        :return: A tuple containing a list of matched entities and their scores.
        """
        matched_entities = []
        matched_entities_scores = []
        for query, query_type in queries.items():
            query = processing_phrases(query)
            if query_type not in self.schema_util.node_en_zh.keys():
                query_type = self.schema_util.get_label_within_prefix(OTHER_TYPE)
            typed_nodes = self.sc.search_vector(
                label=query_type,
                property_key="name",
                query_vector=self.vectorizer.vectorize(query),
                topk=top_k,
            )
            if query_type != self.schema_util.get_label_within_prefix(OTHER_TYPE):
                nontyped_nodes = self.sc.search_vector(
                    label=self.schema_util.get_label_within_prefix(OTHER_TYPE),
                    property_key="name",
                    query_vector=self.vectorizer.vectorize(query),
                    topk=top_k,
                )
            else:
                nontyped_nodes = typed_nodes

            if len(typed_nodes) == 0 and len(nontyped_nodes) != 0:
                matched_entities.append(
                    {"name": nontyped_nodes[0]["node"]["name"], "type": OTHER_TYPE}
                )
                matched_entities_scores.append(nontyped_nodes[0]["score"])
            elif len(typed_nodes) != 0 and len(nontyped_nodes) != 0:
                if typed_nodes[0]["score"] > 0.8:
                    matched_entities.append(
                        {"name": typed_nodes[0]["node"]["name"], "type": query_type}
                    )
                    matched_entities_scores.append(typed_nodes[0]["score"])
                else:
                    matched_entities.append(
                        {"name": nontyped_nodes[0]["node"]["name"], "type": OTHER_TYPE}
                    )
                    matched_entities_scores.append(nontyped_nodes[0]["score"])
                    matched_entities.append(
                        {"name": typed_nodes[0]["node"]["name"], "type": query_type}
                    )
                    matched_entities_scores.append(typed_nodes[0]["score"])
            elif len(typed_nodes) != 0 and len(nontyped_nodes) == 0:
                if typed_nodes[0]["score"] > 0.8:
                    matched_entities.append(
                        {"name": typed_nodes[0]["node"]["name"], "type": query_type}
                    )
                    matched_entities_scores.append(typed_nodes[0]["score"])

        if not matched_entities:
            logger.info(f"No entities matched for {queries}")
        return matched_entities, matched_entities_scores

    def calculate_combined_scores(self, sim_scores: Dict[str, float], pagerank_scores: Dict[str, float]):
        """
        Calculate and return the combined scores that integrate both similarity scores and PageRank scores.

        Parameters:
        sim_scores (Dict[str, float]): A dictionary containing similarity scores, where keys are identifiers and values are scores.
        pagerank_scores (Dict[str, float]): A dictionary containing PageRank scores, where keys are identifiers and values are scores.

        Returns:
        Dict[str, float]: A dictionary containing the combined scores, where keys are identifiers and values are the combined scores.
        """
        def min_max_normalize(x):
            if len(x) == 0:
                return []
            if np.max(x) - np.min(x) > 0:
                return (x - np.min(x)) / (np.max(x) - np.min(x))
            else:
                return x - np.min(x)

        all_keys = set(pagerank_scores.keys()).union(set(sim_scores.keys()))
        for key in all_keys:
            sim_scores.setdefault(key, 0.0)
            pagerank_scores.setdefault(key, 0.0)
        sim_scores = dict(zip(sim_scores.keys(), min_max_normalize(
            np.array(list(sim_scores.values()))
        )))
        pagerank_scores = dict(zip(pagerank_scores.keys(), min_max_normalize(
            np.array(list(pagerank_scores.values()))
        )))
        combined_scores = dict()
        for key in pagerank_scores.keys():
            combined_scores[key] = (sim_scores[key] * (1 - self.pagerank_weight) +
                                    pagerank_scores[key] * self.pagerank_weight
                                    )
        return combined_scores

    def recall_docs(self, query: str, top_k: int = 5, **kwargs):
        """
        Recall relevant documents based on the query string.

        Parameters:
        - query (str): The user's query string.
        - top_k (int, optional): The number of documents to return, default is 5.

        Keyword Arguments:
        - kwargs: Additional keyword arguments.

        Returns:
        - list: A list containing the top_k most relevant documents.
        """
        assert isinstance(query, str), "Query must be a string"

        chunk_nums = top_k * 20
        if chunk_nums == 0:
            return []

        ner_list = self.named_entity_recognition(query)
        print(ner_list)
        if self.with_semantic:
            std_ner_list = self.named_entity_standardization(query, ner_list)
            self.append_official_name(ner_list, std_ner_list)

        entities = {}
        for item in ner_list:
            entity = item.get("entity", "")
            category = item.get("category", "")
            official_name = item.get("official_name", "")
            if not entity or not (category or official_name):
                continue
            if category.lower() in ["works", "person", "other"]:
                entities[entity] = category
            else:
                entities[entity] = official_name or category

        sim_scores = self.calculate_sim_scores(query, chunk_nums)
        matched_entities, matched_scores = self.match_entities(entities)
        pagerank_scores = self.calculate_pagerank_scores(matched_entities)

        if not matched_entities:
            combined_scores = sim_scores
        elif matched_entities and np.min(matched_scores) > self.pagerank_threshold:
            combined_scores = pagerank_scores
        else:
            combined_scores = self.calculate_combined_scores(sim_scores, pagerank_scores)
        sorted_scores = sorted(
            combined_scores.items(), key=lambda item: item[1], reverse=True
        )
        logger.debug(f"sorted_scores: {sorted_scores}")

        return self.get_all_docs_by_id(query, sorted_scores, top_k)

    def get_all_docs_by_id(self, query: str, doc_ids: list, top_k: int):
        """
        Retrieve a list of documents based on their IDs.

        Parameters:
        - query (str): The query string for text matching.
        - doc_ids (list): A list of document IDs to retrieve documents.
        - top_k (int): The maximum number of documents to return.

        Returns:
        - list: A list of matched documents.
        """
        matched_docs = []
        hits_docs = set()
        counter = 0
        for doc_id in doc_ids:
            if counter == top_k:
                break
            if isinstance(doc_id, tuple):
                doc_score = doc_id[1]
                doc_id = doc_id[0]
            else:
                doc_score = doc_ids[doc_id]
            counter += 1
            node = self.reason.query_node(label=self.schema_util.get_label_within_prefix(CHUNK_TYPE), id_value=doc_id)
            node_dict = dict(node.items())
            matched_docs.append(f"#{node_dict['name']}#{node_dict['content']}#{doc_score}")
            hits_docs.add(node_dict['name'])
        try:
            text_matched = self.sc.search_text(query, [self.schema_util.get_label_within_prefix(CHUNK_TYPE)], topk=1)
            if text_matched:
                for item in text_matched:
                    title = item["node"]["name"]
                    if title not in hits_docs:
                        if len(matched_docs) > 0:
                            matched_docs.pop()
                        else:
                            logger.warning(f"{query} matched docs is empty")
                        matched_docs.append(f'#{item["node"]["name"]}#{item["node"]["content"]}#{item["score"]}')
                        break
        except Exception as e:
            logger.warning(f"{query} query chunk failed: {e}", exc_info=True)
        logger.debug(f"matched_docs: {matched_docs}")
        return matched_docs

    def rerank_docs(self, queries: List[str], passages: List[str]):
        """
        Re-ranks the given passages based on the provided queries.

        Parameters:
        - queries (List[str]): A list of queries.
        - passages (List[str]): A list of passages.

        Returns:
        - List[str]: A re-ranked list of passages.
        """
        if self.reranker is None:
            return passages
        return self.reranker.rerank(queries, passages)



from abc import abstractmethod
from typing import Tuple

from kag.interface.solver.lf_planner_abc import LFPlannerABC
from kag.solver.common.base import KagBaseModule
from kag.solver.logic.core_modules.lf_solver import LFSolver


class KagReasonerABC(KagBaseModule):
    """
    A processor class for handling logical form tasks in language processing.

    This class uses an LLM module (llm_module) to plan, retrieve, and solve logical forms.

    Parameters:
    - llm_module: The language model module used for generating and processing logical forms.
    - lf_planner (LFBasePlanner): The planner for structuring logical forms. Defaults to None. If not provided, the default implementation of LFPlanner is used.
    - lf_solver: Instance of the logical form solver, which solves logical form problems.

    Attributes:
    - lf_planner: Instance of the logical form planner.
    - lf_solver: Instance of the logical form solver, which solves logical form problems.
    - sub_query_total: Total number of sub-queries processed.
    - kg_direct: Number of direct knowledge graph queries.
    - trace_log: List to log trace information.
    """
    def __init__(self, lf_planner: LFPlannerABC = None, lf_solver: LFSolver = None, **kwargs):
        super().__init__(**kwargs)

    @abstractmethod
    def reason(self, question: str) -> Tuple[str, str, dict]:
        """
        Processes a given question by planning and executing logical forms to derive an answer.

        Parameters:
        - question (str): The input question to be processed.

        Returns:
        Tuple
        - solved_answer: The final answer derived from solving the logical forms.
        - supporting_fact: Supporting facts gathered during the reasoning process.
        - history_log: A dictionary containing the history of QA pairs and re-ranked documents.
        """


# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import os
from typing import List, Type

from kag.builder.model.chunk import Chunk
from kag.interface.builder import SourceReaderABC
from knext.common.base.runnable import Input, Output


class TXTReader(SourceReaderABC):
    """
    A PDF reader class that inherits from SourceReader.
    """

    @property
    def input_types(self) -> Type[Input]:
        return str

    @property
    def output_types(self) -> Type[Output]:
        return Chunk

    def invoke(self, input: Input, **kwargs) -> List[Output]:
        """
        The main method for processing text reading. This method reads the content of the input (which can be a file path or text content) and converts it into a Chunk object.

        Args:
            input (Input): The input string, which can be the path to a text file or direct text content.
            **kwargs: Additional keyword arguments, currently unused but kept for potential future expansion.

        Returns:
            List[Output]: A list containing Chunk objects, each representing a piece of text read.

        Raises:
            ValueError: If the input is empty.
            IOError: If there is an issue reading the file specified by the input.
        """
        if not input:
            raise ValueError("Input cannot be empty")

        try:
            if os.path.exists(input):
                with open(input, "r", encoding='utf-8') as f:
                    content = f.read()
            else:
                content = input
        except OSError as e:
            raise IOError(f"Failed to read file: {input}") from e

        basename, _ = os.path.splitext(os.path.basename(input))
        chunk = Chunk(
            id=Chunk.generate_hash_id(input),
            name=basename,
            content=content,
        )
        return [chunk]



from typing import Union

from kag.solver.logic.core_modules.common.base_model import LogicNode
from kag.solver.logic.core_modules.common.one_hop_graph import KgGraph
from kag.solver.logic.core_modules.common.schema_utils import SchemaUtils
from kag.solver.logic.core_modules.op_executor.op_executor import OpExecutor
from kag.solver.logic.core_modules.op_executor.op_output.module.get_executor import GetExecutor
from kag.solver.logic.core_modules.parser.logic_node_parser import GetNode
from kag.solver.logic.core_modules.retriver.entity_linker import EntityLinkerBase
from kag.solver.logic.core_modules.retriver.graph_retriver.dsl_executor import DslRunner


class OutputExecutor(OpExecutor):
    def __init__(self, nl_query: str, kg_graph: KgGraph, schema: SchemaUtils, el: EntityLinkerBase, dsl_runner: DslRunner, cached_map: dict, debug_info: dict, **kwargs):
        super().__init__(nl_query, kg_graph, schema, debug_info, **kwargs)
        self.KAG_PROJECT_ID = kwargs.get('KAG_PROJECT_ID')
        self.op_register_map = {
            'get': GetExecutor(nl_query, kg_graph, schema, el, dsl_runner, cached_map, self.debug_info,KAG_PROJECT_ID = kwargs.get('KAG_PROJECT_ID'))
        }

    def is_this_op(self, logic_node: LogicNode) -> bool:
        return isinstance(logic_node, GetNode)

    def executor(self, logic_node: LogicNode, req_id: str, param: dict) -> Union[KgGraph, list]:
        op = self.op_register_map.get(logic_node.operator, None)
        if op is None:
            return []
        return op.executor(logic_node, req_id, param)



# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import io
import json
from pathlib import Path
from abc import ABC, abstractmethod
from typing import Any, Union, Iterable, Tuple

from typing import Dict
import logging

logger = logging.getLogger(__name__)

Item = Dict[str, Any]
RetrievalResult = Iterable[Tuple[Item, float]]


class Retriever(ABC):
    """
    Retriever indexing a collection of items and supports fast retrieving of the
    desired items given a query.
    """

    @classmethod
    def from_config(cls, config: Union[str, Path, Dict[str, Any]]) -> "Retriever":
        """
        Create retriever from `config`.

        If `config` is a string or path, it will be loaded as a dictionary depending
        on its file extension. Currently, the following formats are supported:

        * .json: JSON
        * .json5: JSON with comments support
        * .yaml: YAML

        :param config: retriever config
        :type config: str, Path or Dict[str, Any]
        :return: retriever instance
        :rtype: Retriever
        """
        from kag.common.utils import dynamic_import_class

        if isinstance(config, (str, Path)):
            config_path = config
            if not isinstance(config_path, Path):
                config_path = Path(config_path)
            if config_path.name.endswith(".yaml"):
                import yaml

                with io.open(config_path, "r", encoding="utf-8") as fin:
                    config = yaml.safe_load(fin)
            elif config_path.name.endswith(".json5"):
                import json5

                with io.open(config_path, "r", encoding="utf-8") as fin:
                    config = json5.load(fin)
            elif config_path.name.endswith(".json"):
                with io.open(config_path, "r", encoding="utf-8") as fin:
                    config = json.load(fin)
            else:
                message = "only .json, .json5 and .yaml are supported currently; "
                message += "can not load retriever config from %r" % str(config_path)
                raise RuntimeError(message)
        elif isinstance(config, dict):
            pass
        else:
            message = "only str, Path and dict are supported; "
            message += "invalid retriever config: %r" % (config,)
            raise RuntimeError(message)

        class_name = config.get("retriever")
        if class_name is None:
            message = "retriever class name is not specified"
            raise RuntimeError(message)
        retriever_class = dynamic_import_class(class_name, "retriever")
        if not issubclass(retriever_class, Retriever):
            message = "class %r is not a retriever class" % (class_name,)
            raise RuntimeError(message)
        retriever = retriever_class._from_config(config)
        return retriever

    @classmethod
    @abstractmethod
    def _from_config(cls, config: Dict[str, Any]) -> "Retriever":
        """
        Create retriever from `config`. This method is supposed to be implemented
        by derived classes.

        :param config: retriever config
        :type config: Dict[str, Any]
        :return: retriever instance
        :rtype: Retriever
        """
        message = "abstract method _from_config is not implemented"
        raise NotImplementedError(message)

    def index(self, items: Union[Item, Iterable[Item]]) -> None:
        """
        Add one or more items to the index of the retriever.

        NOTE: This method may not be supported by the retriever.

        :param items: items to index
        :type items: Item or Iterable[Item]
        """
        message = "method index is not supported by the retriever"
        raise RuntimeError(message)

    @abstractmethod
    def retrieve(
            self, queries: Union[str, Iterable[str]], top_k: int = 10
    ) -> Union[RetrievalResult, Iterable[RetrievalResult]]:
        """
        Retrieve items for the given query or queries.

        :param queries: queries to retrieve
        :type queries: str or Iterable[str]
        :param int top_k: how many most related items to return for each query, default to 10
        :return: retrieval results of the queries
        :rtype: RetrievalResult or Iterable[RetrievalResult]
        """
        message = "abstract method retrieve is not implemented"
        raise NotImplementedError(message)





#!/usr/bin/python
# coding=utf8

import csv
import json
from collections import defaultdict
from typing import List

from knext.reasoner.client import ReasonerClient
from knext.schema.model.property import Property
from kag.solver.logic.core_modules.config import LogicFormConfiguration


class SchemaUtils:
    def __init__(self, config: LogicFormConfiguration):
        self.config: LogicFormConfiguration = config
        self.prefix = config.prefix
        self.prefix_concat_sign = "."
        self.project_id = config.project_id
        self.nodes = set()  # zh
        self.edges = set()  # zh
        self.spo = set()  # zh
        self.spo_en = set()  # en
        self.node_attr = {}  # {node_zh:[attr_zh]}
        self.edge_attr = {}
        self.node_edge = {}
        self.so_p = defaultdict(set)
        self.sp_o = defaultdict(set)
        self.op_s = defaultdict(set)
        self.so_p_en = defaultdict(set)
        self.sp_o_en = defaultdict(set)
        self.op_s_en = defaultdict(set)
        self.s_po_en = defaultdict(set)
        self.o_ps_en = defaultdict(set)
        self.node_zh_en = {}
        self.node_en_zh = {}
        self.node_zh_2_full_name = {}
        self.node_en_2_full_name = {}
        self.full_name_2_node_en = {}
        self.edge_zh_en = {}
        self.edge_en_zh = {}
        self.spo_zh_en = {}
        self.spo_en_zh = {}
        self.attr_zh_en = {}
        self.attr_zh_en_by_label = {}
        self.attr_en_zh = {}
        self.attr_en_zh_by_label = {}
        self.attr_enums = {}
        self.ext_attr_name_set = ["basicInfo", "infobox"]
        self.get_schema()

    def get_spo_with_p(self, spo):
        _, p, _ = spo.split('_')
        return p

    def get_label_within_prefix(self, label_name_without_prefix):
        if not label_name_without_prefix:
            return label_name_without_prefix

        if label_name_without_prefix in self.node_en_2_full_name:
            return self.node_en_2_full_name.get(label_name_without_prefix)
        if label_name_without_prefix in self.node_zh_2_full_name:
            return self.node_zh_2_full_name.get(label_name_without_prefix)
        return label_name_without_prefix

    def get_label_without_prefix(self, label_name_with_prefix):
        if not label_name_with_prefix:
            return label_name_with_prefix
        if label_name_with_prefix in self.full_name_2_node_en:
            return self.full_name_2_node_en.get(label_name_with_prefix)
        return label_name_with_prefix

    def _add_attr_with_label(self, label_name, nameZh, name):
        if label_name in self.attr_en_zh_by_label.keys():
            attr_en_zh_tmp = self.attr_en_zh_by_label[label_name]
            attr_en_zh_tmp[name] = nameZh
        else:
            attr_en_zh_tmp = {
                name: nameZh
            }
        self.attr_en_zh_by_label[label_name] = attr_en_zh_tmp

        if label_name in self.attr_zh_en_by_label.keys():
            attr_zh_en_tmp = self.attr_zh_en_by_label[label_name]
            attr_zh_en_tmp[nameZh] = name
        else:
            attr_zh_en_tmp = {
                nameZh: name
            }
        self.attr_zh_en_by_label[label_name] = attr_zh_en_tmp

    def get_attr_en_zh_by_label(self, label_name):
        if label_name not in self.attr_en_zh_by_label.keys():
            return {}
        return self.attr_en_zh_by_label[label_name]

    def get_attr_zh_en_by_label(self, label_name):
        if label_name not in self.attr_zh_en_by_label.keys():
            return {}
        return self.attr_zh_en_by_label[label_name]

    def get_attr(self, label_name, attributes):
        attributes_namezh = []
        if not attributes:
            return attributes_namezh
        for attribute in attributes:
            if not attribute:
                continue
            # print('attribute:', attribute)
            attribute = json.loads(attribute)
            if 'constraints' in attribute and 'name' in attribute['constraints'] and attribute['constraints'][
                'name'] == "Enum":
                enums = list(attribute['constraints']['value'].keys())
            else:
                enums = None
            if attribute['name'].startswith('kg') and attribute['name'].endswith('Raw'):
                continue
            self.attr_zh_en[attribute['nameZh']] = attribute['name']
            self.attr_en_zh[attribute['name']] = attribute['nameZh']
            self.attr_enums[attribute['nameZh']] = enums
            self._add_attr_with_label(label_name, attribute['nameZh'], attribute['name'])
            attributes_namezh.append(attribute['nameZh'])
        return attributes_namezh

    def get_ext_json_prop(self):
        return self.ext_attr_name_set

    def get_schema(self):
        if self.config.schema_file_name == "":
            return self.get_schema_from_spg()
        return self.get_schema_from_csv()

    def get_schema_from_spg(self):
        sc = ReasonerClient(self.config.host_addr, int(self.project_id))
        spg_schema = sc.get_reason_schema()
        node_attributes = {}
        for spg_type_name, spg_type in spg_schema.items():
            name_en = spg_type.name_en
            name_zh = spg_type.name_zh
            self.node_en_2_full_name[name_en] = spg_type.name
            self.full_name_2_node_en[spg_type.name] = name_en
            self.node_zh_2_full_name[name_zh] = spg_type.name
            father_en = spg_type.parent_type_name
            attributes = self._convert_spg_attr_set(list(spg_type.properties.values()))
            if father_en and father_en in node_attributes:
                attributes += node_attributes[father_en]
            node_attributes[name_en] = attributes

            self.nodes.add(name_zh)
            self.node_zh_en[name_zh] = name_en
            self.node_en_zh[name_en] = name_zh
            entity_default_attributes = [
                '{"name": "name", "nameZh": "名称"}',
                '{"name": "id", "nameZh": "实体主键"}',
                '{"name": "description", "nameZh": "描述"}'
            ]
            attributes += entity_default_attributes
            attributes_namezh = self.get_attr(name_en, attributes)
            self.node_attr[name_zh] = attributes_namezh

            relations = spg_type.relations.values()
            for relation in relations:
                s_name_en = name_en
                s_name_zh = name_zh
                p_name_en = relation.name
                p_name_zh = relation.name_zh
                spg_o_type = spg_schema[relation.object_type_name]
                o_name_en = spg_o_type.name_en
                o_name_zh = spg_o_type.name_zh
                r_name_zh = f"{s_name_zh}_{p_name_zh}_{o_name_zh}"
                r_name_en = f"{s_name_en}_{p_name_en}_{o_name_en}"

                if r_name_zh not in self.spo:
                    self.spo.add(r_name_zh)
                if r_name_en not in self.spo_en:
                    self.spo_en.add(r_name_en)

                self.spo_en_zh[r_name_en] = r_name_zh
                self.spo_zh_en[r_name_zh] = r_name_en
                self.so_p[(s_name_zh, o_name_zh)].add(p_name_zh)
                self.sp_o[(s_name_zh, p_name_zh)].add(o_name_zh)
                self.op_s[(o_name_zh, p_name_zh)].add(s_name_zh)
                self.so_p_en[(s_name_en, o_name_en)].add(p_name_en)
                self.sp_o_en[(s_name_en, p_name_en)].add(o_name_en)
                self.op_s_en[(o_name_en, p_name_en)].add(s_name_en)
                self.s_po_en[s_name_en].add((p_name_en, o_name_en))
                self.o_ps_en[o_name_en].add((p_name_en, s_name_en))
                self.edges.add(r_name_zh)
                self.edge_zh_en[r_name_zh] = r_name_en
                self.edge_en_zh[r_name_en] = r_name_zh
                if s_name_zh not in self.node_edge:
                    self.node_edge[s_name_zh] = set()
                self.node_edge[s_name_zh].add(p_name_zh)
                if o_name_zh not in self.node_edge:
                    self.node_edge[o_name_zh] = set()
                self.node_edge[o_name_zh].add(p_name_zh)
                r_attributes = self._convert_spg_attr_set(list(relation.sub_properties.values()))
                r_attributes_namezh = self.get_attr(name_en, r_attributes)
                self.edge_attr[p_name_zh] = r_attributes_namezh

    def get_schema_from_csv(self):
        f = open(self.config.schema_file_name)

        reader = csv.reader(f)
        next(reader)
        # next(reader)
        node_attributes = {}
        for row in reader:
            obj, name_zh, name_en, father_en, edge_direction, attributes = row[0], row[1], row[2], row[3], row[4], row[
                                                                                                                   6:]
            if "nodeType/edgeType" in obj:
                continue
            name_en = name_en.replace(self.prefix, '')
            # if name_en in ['Event', 'ProductTaxon']:   
            if father_en and father_en in node_attributes:
                attributes += node_attributes[father_en]
            node_attributes[name_en] = attributes
            if obj not in ['edge', 'inputEdge']:
                # if name_zh in ['百科实体', '热点事件', '事件']:
                if name_zh in ['百科实体']:
                    continue
                self.nodes.add(name_zh)
                self.node_zh_en[name_zh] = name_en
                self.node_en_zh[name_en] = name_zh
                entity_default_attributes = [
                    '{"name": "name", "nameZh": "名称"}',
                    '{"name": "id", "nameZh": "实体主键"}',
                    '{"name": "description", "nameZh": "描述"}'
                ]
                attributes += entity_default_attributes
                attributes_namezh = self.get_attr(name_en, attributes)
                self.node_attr[name_zh] = attributes_namezh


            elif obj == 'edge':
                s, p, o = name_zh.split('_')
                # if s in ['百科实体', '热点事件', '事件'] or o in ['百科实体', '热点事件', '事件']:
                if s in ['百科实体'] or o in ['百科实体']:
                    continue
                if name_zh not in self.spo:
                    self.spo.add(name_zh)
                if name_en not in self.spo_en:
                    self.spo_en.add(name_en)
                self.spo_en_zh[name_en] = name_zh
                self.spo_zh_en[name_zh] = name_en
                self.so_p[(s, o)].add(p)
                self.sp_o[(s, p)].add(o)
                self.sp_o[(o, p)].add(s)
                s_en, p_en, o_en = name_en.split('_')
                self.so_p_en[(s_en, o_en)].add(p_en)
                self.sp_o_en[(s_en, p_en)].add(o_en)
                self.op_s_en[(o_en, p_en)].add(s_en)
                self.s_po_en[s_en].add((p_en, o_en))
                self.o_ps_en[o_en].add((p_en, s_en))
                self.edges.add(p)
                self.edge_zh_en[p] = p_en
                self.edge_en_zh[p_en] = p
                if s not in self.node_edge:
                    self.node_edge[s] = set()
                self.node_edge[s].add(p)
                if o not in self.node_edge:
                    self.node_edge[o] = set()
                self.node_edge[o].add(p)
                attributes_namezh = self.get_attr(name_en, attributes)
                self.edge_attr[p] = attributes_namezh

    def get_schema_rdf(self, path_node, path_edge):
        f_node = open(path_node)
        f_edge = open(path_edge)
        for row in csv.DictReader(f_node):
            name, id = row['name'], row['alias']
            self.nodes.add(name)
        for row in csv.DictReader(f_edge):
            name = row['name']
            self.edges.add(name)

    def _convert_spg_attr_set(self, attr_set: List[Property]):
        return [json.dumps({
            'constraints': attr.to_dict().get('constraint', {}),
            'name': attr.to_dict().get('name'),
            'nameZh': attr.to_dict().get('name_zh')
        }) for attr in attr_set]


def generate_nodes_edges_hetero(schema):
    '''
    nodes {
        hetero {
            "CommonSenseKG.Person" {
                fe: [
                    "gender;Raw|use_fe=False;Direct;str",
                ]
            }
            "CommonSenseKG.Work" {
                fe: [
                    "releaseDate;Raw|use_fe=False;Direct;str",
                ]
            }
        }
    }
    edges {
        hetero {
            "CommonSenseKG.Person_debutWork_CommonSenseKG.Work" {
                fe: []
            }
            "CommonSenseKG.Person_representativeWork_CommonSenseKG.Work" {
                fe: []
            }
        }
    }
    '''
    nodes_hetero, edges_hetero = defaultdict(dict), defaultdict(dict)
    for node in schema.nodes:
        node = schema.node_zh_en[node]
        features = []
        for attr in schema.node_attr[schema.node_en_zh[node]]:
            attr = schema.attr_zh_en[attr]
            features.append(attr + ';Raw|use_fe=False;Direct;str')
        node = schema.prefix + '.' + node
        nodes_hetero[node] = {'fe': features}

    for spo in schema.spo_en:
        s, p, o = spo.split('_')
        edge = '_'.join([schema.prefix + '.' + s, p, schema.prefix + '.' + o])
        edges_hetero[edge] = {'fe': []}

    print('nodes_hetero:', json.dumps(nodes_hetero, indent=2).replace('"fe"', 'fe').replace('},', '}'))
    print('edges_hetero:', json.dumps(edges_hetero, indent=2).replace('"fe"', 'fe').replace('},', '}'))



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import json
from typing import Optional, List, Dict, Any

from kag.common.base.prompt_op import PromptOp


class OpenIETriplePrompt(PromptOp):

    template_zh = """
{
    "instruction": "您是一位专门从事开放信息提取（OpenIE）的专家。请从input字段的文本中提取任何可能的关系（包括主语、谓语、宾语），并按照JSON格式列出它们，须遵循example字段的示例格式。请注意以下要求：1. 每个三元组应至少包含entity_list实体列表中的一个，但最好是两个命名实体。2. 明确地将代词解析为特定名称，以保持清晰度。",
    "entity_list": $entity_list,
    "input": "$input",
    "example": {
        "input": "烦躁不安、语妄、失眠酌用镇静药，禁用抑制呼吸的镇静药。\n3.并发症的处理经抗菌药物治疗后，高热常在24小时内消退，或数日内逐渐下降。\n若体温降而复升或3天后仍不降者，应考虑SP的肺外感染，如腋胸、心包炎或关节炎等。治疗：接胸腔压力调节管＋吸引机负压吸引水瓶装置闭式负压吸引宜连续，如经12小时后肺仍未复张，应查找原因。",
        "entity_list": [
            {"entity": "烦躁不安", "category": "Symptom"},
            {"entity": "语妄", "category": "Symptom"},
            {"entity": "失眠", "category": "Symptom"},
            {"entity": "镇静药", "category": "Medicine"},
            {"entity": "肺外感染", "category": "Disease"},
            {"entity": "胸腔压力调节管", "category": "MedicalEquipment"},
            {"entity": "吸引机负压吸引水瓶装置", "category": "MedicalEquipment"},
            {"entity": "闭式负压吸引", "category": "SurgicalOperation"}
        ],
        "output":[
            ["烦躁不安", "酌用", "镇静药"],
            ["语妄", "酌用", "镇静药"],
            ["失眠", "酌用", "镇静药"],
            ["镇静药", "禁用", "抑制呼吸的镇静药"],
            ["高热", "消退", "24小时内"],
            ["高热", "下降", "数日内"],
            ["体温", "降而复升或3天后仍不降", "肺外感染"],
            ["肺外感染", "考虑", "腋胸、心包炎或关节炎"],
            ["胸腔压力调节管", "接", "吸引机负压吸引水瓶装置"],
            ["闭式负压吸引", "宜连续", "如经12小时后肺仍未复张"]
        ]
    }
}    
    """

    template_en = template_zh

    def __init__(self, language: Optional[str] = "en"):
        super().__init__(language)

    @property
    def template_variables(self) -> List[str]:
        return ["entity_list", "input"]

    def parse_response(self, response: str, **kwargs):
        rsp = response
        if isinstance(rsp, str):
            rsp = json.loads(rsp)
        if isinstance(rsp, dict) and "output" in rsp:
            rsp = rsp["output"]
        if isinstance(rsp, dict) and "triples" in rsp:
            triples = rsp["triples"]
        else:
            triples = rsp

        standardized_triples = []
        for triple in triples:
            if isinstance(triple, list):
                standardized_triples.append(triple)
            elif isinstance(triple, dict):
                s = triple.get("subject")
                p = triple.get("predicate")
                o = triple.get("object")
                if s and p and o:
                    standardized_triples.append([s, p, o])

        return standardized_triples
